#!/usr/bin/env python3

r'''Simulates different chessboard dances to find the best technique

We want the shortest chessboard dances that produce the most confident results.
In a perfect world would put the chessboard in all the locations where we would
expect to use the visual system, since the best confidence is obtained in
regions where the chessboard was observed.

However, due to geometric constraints it is sometimes impossible to put the
board in the right locations. This tool clearly shows that filling the field of
view produces best results. But very wide lenses require huge chessboards,
displayed very close to the lens in order to fill the field of view. This means
that using a wide lens to look out to infinity will always result in potentially
too little projection confidence. This tool is intended to find the kind of
chessboard dance to get good confidences by simulating different geometries and
dances.

We arrange --Ncameras cameras horizontally, with an identity rotation, evenly
spaced with a spacing of --camera-spacing meters. The left camera is at the
origin.

We show the cameras lots of dense chessboards ("dense" meaning that every camera
sees all the points of all the chessboards). The chessboard come in two
clusters: "near" and "far". Each cluster is centered straight ahead of the
midpoint of all the cameras, with some random noise on the position and
orientation. The distances from the center of the cameras to the center of the
clusters are given by --range. This tool solves the calibration problem, and
generates uncertainty-vs-range curves. Each run of this tool generates a family
of this curve, for different values of Nframes-far, the numbers of chessboard
observations in the "far" cluster.

This tool scans some parameter (selected by --scan), and reports the
uncertainty-vs-range for the different values of this parameter (as a plot and
as an output vnl).

If we don't want to scan any parameter, and just want a single
uncertainty-vs-range plot, don't pass --scan.

'''

import sys
import argparse
import re
import os

def parse_args():

    def positive_float(string):
        try:
            value = float(string)
        except:
            raise argparse.ArgumentTypeError("argument MUST be a positive floating-point number. Got '{}'".format(string))
        if value <= 0:
            raise argparse.ArgumentTypeError("argument MUST be a positive floating-point number. Got '{}'".format(string))
        return value
    def positive_int(string):
        try:
            value = int(string)
        except:
            raise argparse.ArgumentTypeError("argument MUST be a positive integer. Got '{}'".format(string))
        if value <= 0 or abs(value-float(string)) > 1e-6:
            raise argparse.ArgumentTypeError("argument MUST be a positive integer. Got '{}'".format(string))
        return value

    parser = \
        argparse.ArgumentParser(description = __doc__,
                                formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('--icam-uncertainty',
                        default = 0,
                        type=int,
                        help='''Which camera to use for the uncertainty reporting. I use the left-most one
                        (camera 0) by default''')
    parser.add_argument('--camera-spacing',
                        default = 0.3,
                        type=float,
                        help='How many meters between adjacent cameras in our synthetic world')
    parser.add_argument('--object-spacing',
                        default="0.077",
                        type=str,
                        help='''Width of each square in the calibration board, in meters. If --scan
                        object_width_n, this is used as the spacing for the
                        SMALLEST width. As I change object_width_n, I adjust
                        object_spacing accordingly, to keep the total board size
                        constant. If --scan object_spacing, this is MIN,MAX to
                        specify the bounds of the scan. In that case
                        object_width_n stays constant, so the board changes
                        size''')
    parser.add_argument('--object-width-n',
                        type=str,
                        help='''How many points the calibration board has per horizontal side. If both are
                        omitted, we default the width and height to 10 (both
                        must be specified to use a non-default value). If --scan
                        object_width_n, this is MIN,MAX to specify the bounds of
                        the scan. In that case I assume a square object, and
                        ignore --object-height-n. Scanning object-width-n keeps
                        the board size constant''')
    parser.add_argument('--object-height-n',
                        type=int,
                        help='''How many points the calibration board has per vertical side. If both are
                        omitted, we default the width and height to 10 (both
                        must be specified to use a non-default value). If --scan
                        object_width_n, this is ignored, and set equivalent to
                        the object width''')
    parser.add_argument('--observed-pixel-uncertainty',
                        type=positive_float,
                        default = 1.0,
                        help='''The standard deviation of x and y pixel coordinates of the input observations
                        I generate. The distribution of the inputs is gaussian,
                        with the standard deviation specified by this argument.
                        Note: this is the x and y standard deviation, treated
                        independently. If each of these is s, then the LENGTH of
                        the deviation of each pixel is a Rayleigh distribution
                        with expected value s*sqrt(pi/2) ~ s*1.25''')
    parser.add_argument('--lensmodel',
                        required=False,
                        type=str,
                        help='''Which lens model to use for the simulation. If omitted, we use the model
                        given on the commandline. We may want to use a
                        parametric model to generate data (model on the
                        commandline), but a richer splined model to solve''')
    parser.add_argument('--skip-calobject-warp-solve',
                        action='store_true',
                        default=False,
                        help='''By default we assume the calibration target is
                        slightly deformed, and we compute this deformation. If
                        we want to assume that the chessboard shape is fixed,
                        pass this option. The actual shape of the board is given
                        by --calobject-warp''')
    parser.add_argument('--calobject-warp',
                        type=float,
                        nargs=2,
                        default=(0.002, -0.005),
                        help='''The "calibration-object warp". These specify the
                        flex of the chessboard. By default, the board is
                        slightly warped (as is usually the case in real life).
                        To use a perfectly flat board, specify "0 0" here''')

    parser.add_argument('--show-geometry-first-solve',
                        action = 'store_true',
                        help='''If given, display the camera, chessboard geometry after the first solve, and
                        exit. Used for debugging''')
    parser.add_argument('--show-uncertainty-first-solve',
                        action = 'store_true',
                        help='''If given, display the uncertainty (at infinity) and observations after the
                        first solve, and exit. Used for debugging''')
    parser.add_argument('--write-models-first-solve',
                        action = 'store_true',
                        help='''If given, write the solved models to disk after the first solve, and exit.
                        Used for debugging. Useful to check fov_x_deg when solving for a splined model''')

    parser.add_argument('--which',
                        choices=('all-cameras-must-see-full-board',
                                 'some-cameras-must-see-full-board',
                                 'all-cameras-must-see-half-board',
                                 'some-cameras-must-see-half-board'),
                        default='all-cameras-must-see-half-board',
                        help='''What kind of random poses are accepted. Default
                        is all-cameras-must-see-half-board''')

    parser.add_argument('--fixed-frames',
                        action='store_true',
                        help='''Optimize the geometry keeping the chessboard frames fixed. This reduces the
                        freedom of the solution, and produces more confident
                        calibrations. It's possible to use this in reality by
                        surverying the chessboard poses''')
    parser.add_argument('--method',
                        choices=('cross-reprojection-ccp', 'mean-pcam', 'cross-reprojection-rrp-Jfp'),
                        default='cross-reprojection-ccp',
                        help='''Multiple uncertainty quantification methods are
                        available. We default to 'cross-reprojection-ccp'; most
                        people should not touch this''')
    parser.add_argument('--ymax',
                        type=float,
                        default = 10.0,
                        help='''If given, use this as the upper extent of the uncertainty plot.''')
    parser.add_argument('--uncertainty-at-range-sampled-min',
                        type=float,
                        help='''If given, use this as the lower bound of the ranges we look at when
                        evaluating projection confidence''')
    parser.add_argument('--uncertainty-at-range-sampled-max',
                        type=float,
                        help='''If given, use this as the upper bound of the ranges we look at when
                        evaluating projection confidence''')
    parser.add_argument('--explore',
                        action='store_true',
                        help='''Drop into a REPL at the end''')

    parser.add_argument('--scan',
                        type=str,
                        default='',
                        choices=('range',
                                 'tilt_radius',
                                 'Nframes',
                                 'Ncameras',
                                 'object_width_n',
                                 'object_spacing',
                                 'num_far_constant_Nframes_near',
                                 'num_far_constant_Nframes_all'),
                        help='''Study the effect of some parameter on uncertainty. The parameter is given as
                        an argument ot this function. Valid choices:
                        ('range','tilt_radius','Nframes','Ncameras',
                        'object_width_n', 'object_spacing', 'num_far_constant_Nframes_near',
                        'num_far_constant_Nframes_all'). Scanning object-width-n
                        keeps the board size constant''')
    parser.add_argument('--scan-object-spacing-compensate-range-from',
                        type=float,
                        help=f'''Only applies if --scan object-spacing. By
                        default we vary the object spacings without touching
                        anything else: the chessboard grows in size as we
                        increase the spacing. If given, we try to keep the
                        apparent size constant: as the object spacing grows, we
                        increase the range. The nominal range given in this
                        argument''')

    parser.add_argument('--range',
                        default = '0.5,4.0',
                        type=str,
                        help='''if --scan num_far_...: this is "NEAR,FAR"; specifying the near and far ranges
                        to the chessboard to evaluate. if --scan range this is
                        "MIN,MAX"; specifying the extents of the ranges to
                        evaluate. Otherwise: this is RANGE, the one range of
                        chessboards to evaluate''')
    parser.add_argument('--tilt-radius',
                        default='30.',
                        type=str,
                        help='''The radius of the uniform distribution used to
                        sample the pitch and yaw of the chessboard
                        observations, in degrees. The default is 30, meaning
                        that the chessboard pitch and yaw are sampled from
                        [-30deg,30deg]. if --scan tilt_radius: this is
                        TILT-MIN,TILT-MAX specifying the bounds of the two
                        tilt-radius values to evaluate. Otherwise: this is the
                        one value we use''')
    parser.add_argument('--roll-radius',
                        type=float,
                        default=20.,
                        help='''The synthetic chessboard orientation is sampled
                        from a uniform distribution: [-RADIUS,RADIUS]. The
                        pitch,yaw radius is specified by the --tilt-radius. The
                        roll radius is selected here. "Roll" is the rotation
                        around the axis normal to the chessboard plane. Default
                        is 20deg''')
    parser.add_argument('--x-radius',
                        type=float,
                        default=None,
                        help='''The synthetic chessboard position is sampled
                        from a uniform distribution: [-RADIUS,RADIUS]. The x
                        radius is selected here. "x" is direction in the
                        chessboard plane, and is also the axis along which the
                        cameras are distributed. A resonable default (possibly
                        range-dependent) is chosen if omitted. MUST be None if
                        --scan num_far_...''')
    parser.add_argument('--y-radius',
                        type=float,
                        default=None,
                        help='''The synthetic chessboard position is sampled
                        from a uniform distribution: [-RADIUS,RADIUS]. The y
                        radius is selected here. "y" is direction in the
                        chessboard plane, and is also normal to the axis along
                        which the cameras are distributed. A resonable default
                        (possibly range-dependent) is chosen if omitted. MUST be
                        None if --scan num_far_...''')
    parser.add_argument('--z-radius',
                        type=float,
                        default=None,
                        help='''The synthetic chessboard position is sampled
                        from a uniform distribution: [-RADIUS,RADIUS]. The z
                        radius is selected here. "z" is direction normal to the
                        chessboard plane. A resonable default (possibly
                        range-dependent) is chosen if omitted. MUST be None if
                        --scan num_far_...''')
    parser.add_argument('--Ncameras',
                        default = '1',
                        type=str,
                        help='''How many cameras oberve our synthetic world. By default we just have one
                        camera. if --scan Ncameras: this is
                        NCAMERAS-MIN,NCAMERAS-MAX specifying the bounds of the
                        camera counts to evaluate. Otherwise: this is the one
                        Ncameras value to use''')
    parser.add_argument('--Nframes',
                        type=str,
                        help='''How many observed frames we have. Ignored if --scan num_far_... if
                        --scan Nframes: this is NFRAMES-MIN,NFRAMES-MAX specifying the bounds of the
                        frame counts to evaluate. Otherwise: this is the one Nframes value to use''')

    parser.add_argument('--Nframes-near',
                        type=positive_int,
                        help='''Used if --scan num_far_constant_Nframes_near. The number of "near" frames is
                        given by this argument, while we look at the effect of
                        adding more "far" frames.''')
    parser.add_argument('--Nframes-all',
                        type=positive_int,
                        help='''Used if --scan num_far_constant_Nframes_all. The number of "near"+"far"
                        frames is given by this argument, while we look at the
                        effect of replacing "near" frames with "far" frames.''')

    parser.add_argument('--Nscan-samples',
                        type=positive_int,
                        default=8,
                        help='''How many values of the parameter being scanned to evaluate. If we're scanning
                        something (--scan ... given) then by default the scan
                        evaluates 8 different values. Otherwise this is set to 1''')

    parser.add_argument('--hardcopy',
                        help=f'''Filename to plot into. If omitted, we make an interactive plot. This is
                        passed directly to gnuplotlib''')
    parser.add_argument('--terminal',
                        help=f'''gnuplot terminal to use for the plots. This is passed directly to
                        gnuplotlib. Omit this unless you know what you're doing''')
    parser.add_argument('--extratitle',
                        help=f'''Extra title string to add to a plot''')
    parser.add_argument('--title',
                        help=f'''Full title string to use in a plot. Overrides
                        the default and --extratitle''')
    parser.add_argument('--set',
                        type=str,
                        action='append',
                        help='''Extra 'set' directives to gnuplotlib. Can be given multiple times''')
    parser.add_argument('--unset',
                        type=str,
                        action='append',
                        help='''Extra 'unset' directives to gnuplotlib. Can be given multiple times''')

    parser.add_argument('model',
                        type = str,
                        help='''Baseline camera model. I use the intrinsics from this model to generate
                        synthetic data. We probably want the "true" model to not
                        be too crazy, so this should probably by a parametric
                        (not splined) model''')

    args = parser.parse_args()
    if args.title is not None and args.extratitle is not None:
        print("--title and --extratitle are mutually exclusive", file=sys.stderr)
        sys.exit(1)

    return args

args = parse_args()

if args.object_width_n  is None and \
   args.object_height_n is None:
    args.object_width_n  = "10"
    args.object_height_n = 10
elif not ( args.object_width_n  is not None and \
           args.object_height_n is not None) and \
           args.scan != 'object_width_n':
    raise Exception("Either --object-width-n or --object-height-n are given: you must pass both or neither")

# arg-parsing is done before the imports so that --help works without building
# stuff, so that I can generate the manpages and README


import numpy as np
import numpysane as nps
import gnuplotlib as gp
import copy
import os.path

# I import the LOCAL mrcal
scriptdir = os.path.dirname(os.path.realpath(__file__))
sys.path[:0] = f"{scriptdir}/../..",
import mrcal


def split_list(s, t):
    r'''Splits a comma-separated list

    None    -> None
    "A,B,C" -> (A,B,C)
    "A"     -> A
    '''
    if s is None: return None,0
    l = [t(x) for x in s.split(',')]
    if len(l) == 1: return l[0],1
    return l,len(l)

def first(s):
    r'''first in a list, or the value if a scalar'''
    if hasattr(s, '__iter__'): return s[0]
    return s

def last(s):
    r'''last in a list, or the value if a scalar'''
    if hasattr(s, '__iter__'): return s[-1]
    return s



controllable_args          = \
    dict( tilt_radius    = dict(_type = float),
          range          = dict(_type = float),
          Ncameras       = dict(_type = int),
          Nframes        = dict(_type = int),
          object_spacing = dict(_type = float),
          object_width_n = dict(_type = int) )

for a in controllable_args.keys():
    l,n = split_list(getattr(args, a), controllable_args[a]['_type'])
    controllable_args[a]['value']   = l
    controllable_args[a]['listlen'] = n

if any( controllable_args[a]['listlen'] > 2 for a in controllable_args.keys() ):
    raise Exception(f"All controllable args must have either at most 2 values. {a} has {controllable_args[a]['listlen']}")

controllable_arg_0values = [ a for a in controllable_args.keys() if controllable_args[a]['listlen'] == 0 ]
controllable_arg_2values = [ a for a in controllable_args.keys() if controllable_args[a]['listlen'] == 2 ]

if    len(controllable_arg_2values) == 0: controllable_arg_2values = ''
elif  len(controllable_arg_2values) == 1: controllable_arg_2values = controllable_arg_2values[0]
else: raise Exception(f"At most 1 controllable arg may have 2 values. Instead I saw: {controllable_arg_2values}")

if re.match("num_far_constant_Nframes_", args.scan):

    if args.x_radius is not None or \
       args.y_radius is not None or \
       args.z_radius is not None:
        raise Exception("--x-radius and --y-radius and --z-radius are exclusive with --scan num_far_...")

    # special case
    if 'Nframes' not in controllable_arg_0values:
        raise Exception(f"I'm scanning '{args.scan}', so --Nframes must not have been given")
    if 'range' != controllable_arg_2values:
        raise Exception(f"I'm scanning '{args.scan}', so --range must have 2 values")
    if args.scan == "num_far_constant_Nframes_near":
        if args.Nframes_all is not None:
            raise Exception(f"I'm scanning '{args.scan}', so --Nframes-all must not have have been given")
        if args.Nframes_near is None:
            raise Exception(f"I'm scanning '{args.scan}', so --Nframes-near must have have been given")
    else:
        if args.Nframes_near is not None:
            raise Exception(f"I'm scanning '{args.scan}', so --Nframes-near must not have have been given")
        if args.Nframes_all is None:
            raise Exception(f"I'm scanning '{args.scan}', so --Nframes-all must have have been given")

else:
    if args.scan != controllable_arg_2values:
        # This covers the scanning-nothing-no2value-anything case
        raise Exception(f"I'm scanning '{args.scan}', the arg given 2 values is '{controllable_arg_2values}'. They must match")
    if len(controllable_arg_0values):
        raise Exception(f"I'm scanning '{args.scan}', so all controllable args should have some value. Missing: '{controllable_arg_0values}")
    if args.scan == '':
        args.Nscan_samples = 1





Nuncertainty_at_range_samples = 80
uncertainty_at_range_sampled_min = args.uncertainty_at_range_sampled_min
if uncertainty_at_range_sampled_min is None:
    uncertainty_at_range_sampled_min = first(controllable_args['range']['value'])/10.
uncertainty_at_range_sampled_max = args.uncertainty_at_range_sampled_max
if uncertainty_at_range_sampled_max is None:
    uncertainty_at_range_sampled_max = last(controllable_args['range']['value']) *10.

uncertainty_at_range_samples = \
    np.logspace( np.log10(uncertainty_at_range_sampled_min),
                 np.log10(uncertainty_at_range_sampled_max),
                 Nuncertainty_at_range_samples)

# I want the RNG to be deterministic
np.random.seed(0)

model_intrinsics = mrcal.cameramodel(args.model)

calobject_warp_true_ref = np.array(args.calobject_warp)


def solve(Ncameras,
          Nframes_near, Nframes_far,
          object_spacing,
          models_true,

          # q.shape            = (Nframes, Ncameras, object_height, object_width, 2)
          # Rt_ref_board.shape = (Nframes, 4,3)

          q_true_near, Rt_ref_board_true_near,
          q_true_far,  Rt_ref_board_true_far,
          fixed_frames = args.fixed_frames):

    q_true_near            = q_true_near            [:Nframes_near]
    Rt_ref_board_true_near = Rt_ref_board_true_near[:Nframes_near]

    if q_true_far is not None:
        q_true_far             = q_true_far             [:Nframes_far ]
        Rt_ref_board_true_far  = Rt_ref_board_true_far [:Nframes_far ]
    else:
        q_true_far             = np.zeros( (0,) + q_true_near.shape[1:],
                                            dtype = q_true_near.dtype)
        Rt_ref_board_true_far  = np.zeros( (0,) + Rt_ref_board_true_near.shape[1:],
                                            dtype = Rt_ref_board_true_near.dtype)

    calobject_warp_true = calobject_warp_true_ref.copy()

    Nframes_all = Nframes_near + Nframes_far

    Rt_ref_board_true = nps.glue( Rt_ref_board_true_near,
                                  Rt_ref_board_true_far,
                                  axis = -3 )

    # Dense observations. All the cameras see all the boards
    indices_frame_camera = np.zeros( (Nframes_all*Ncameras, 2), dtype=np.int32)
    indices_frame = indices_frame_camera[:,0].reshape(Nframes_all,Ncameras)
    indices_frame.setfield(nps.outer(np.arange(Nframes_all, dtype=np.int32),
                                     np.ones((Ncameras,), dtype=np.int32)),
                           dtype = np.int32)
    indices_camera = indices_frame_camera[:,1].reshape(Nframes_all,Ncameras)
    indices_camera.setfield(nps.outer(np.ones((Nframes_all,), dtype=np.int32),
                                     np.arange(Ncameras, dtype=np.int32)),
                           dtype = np.int32)
    indices_frame_camintrinsics_camextrinsics = \
        nps.glue(indices_frame_camera,
                 indices_frame_camera[:,(1,)],
                 axis=-1)

    # If not fixed_frames: we use camera0 as the reference cordinate system, and
    # we allow the chessboard poses to move around. Else: the reference
    # coordinate system is arbitrary, but all cameras are allowed to move
    # around. The chessboards poses are fixed
    if not fixed_frames:
        indices_frame_camintrinsics_camextrinsics[:,2] -= 1

    q = nps.glue( q_true_near,
                  q_true_far,
                  axis = -5 )

    # apply noise
    q += np.random.randn(*q.shape) * args.observed_pixel_uncertainty

    # The observations are dense (in the data every camera sees all the
    # chessboards), but some of the observations WILL be out of bounds. I
    # pre-mark those as outliers so that the solve doesn't do weird stuff

    # Set the weights to 1 initially
    # shape (Nframes, Ncameras, object_height_n, object_width_n, 3)
    observations = nps.glue(q,
                            np.ones( q.shape[:-1] + (1,) ),
                            axis = -1)

    # shape (Ncameras, 1, 1, 2)
    imagersizes = nps.mv( nps.cat(*[ m.imagersize() for m in models_true ]),
                          -2, -4 )

    # mark the out-of-view observations as outliers
    observations[ np.any( q              < 0, axis=-1 ), 2 ] = -1.
    observations[ np.any( q-imagersizes >= 0, axis=-1 ), 2 ] = -1.

    # shape (Nobservations, Nh, Nw, 2)
    observations = nps.clump( observations,
                   n = 2 )

    intrinsics = nps.cat( *[m.intrinsics()[1]         for m in models_true]     )

    # If not fixed_frames: we use camera0 as the reference cordinate system, and
    # we allow the chessboard poses to move around. Else: the reference
    # coordinate system is arbitrary, but all cameras are allowed to move
    # around. The chessboards poses are fixed
    if fixed_frames:
        extrinsics = nps.cat( *[m.rt_cam_ref() for m in models_true] )
    else:
        extrinsics = nps.cat( *[m.rt_cam_ref() for m in models_true[1:]] )
    if len(extrinsics) == 0: extrinsics = None

    if nps.norm2(models_true[0].rt_cam_ref()) > 1e-6:
        raise Exception("models_true[0] must sit at the origin")
    imagersizes = nps.cat( *[m.imagersize() for m in models_true] )


    optimization_inputs = \
        dict( # intrinsics filled in later
              rt_cam_ref                                = extrinsics,
              rt_ref_frame                              = mrcal.rt_from_Rt(Rt_ref_board_true),
              points                                    = None,
              observations_board                        = observations,
              indices_frame_camintrinsics_camextrinsics = indices_frame_camintrinsics_camextrinsics,
              observations_point                        = None,
              indices_point_camintrinsics_camextrinsics = None,
              # lensmodel filled in later
              calobject_warp                            = copy.deepcopy(calobject_warp_true),
              imagersizes                               = imagersizes,
              calibration_object_spacing                = object_spacing,
              verbose                                   = False,
              # do_optimize_extrinsics filled in later
              # do_optimize_intrinsics_core filled in later
              do_optimize_frames                        = False,
              do_optimize_intrinsics_distortions        = True,
              do_optimize_calobject_warp                = False, # turn this on, and reoptimize later, if needed
              do_apply_regularization                   = True,
              do_apply_outlier_rejection                = False)


    if args.lensmodel is None:
        lensmodel = model_intrinsics.intrinsics()[0]
    else:
        lensmodel = args.lensmodel
    Nintrinsics = mrcal.lensmodel_num_params(lensmodel)

    if re.search("SPLINED", lensmodel):

        # These are already mostly right, So I lock them down while I seed the
        # intrinsics
        optimization_inputs['do_optimize_extrinsics'] = False

        # I pre-optimize the core, and then lock it down
        optimization_inputs['lensmodel']                   = 'LENSMODEL_STEREOGRAPHIC'
        optimization_inputs['intrinsics']                  = intrinsics[:,:4].copy()
        optimization_inputs['do_optimize_intrinsics_core'] = True

        stats = mrcal.optimize(**optimization_inputs)
        print(f"## optimized. rms = {stats['rms_reproj_error__pixels']}", file=sys.stderr)

        # core is good. Lock that down, and get an estimate for the control
        # points
        optimization_inputs['do_optimize_intrinsics_core'] = False
        optimization_inputs['lensmodel']                   = lensmodel
        optimization_inputs['intrinsics']                  = nps.glue(optimization_inputs['intrinsics'],
                                                                      np.zeros((Ncameras,Nintrinsics-4),),axis=-1)
        stats = mrcal.optimize(**optimization_inputs)
        print(f"## optimized. rms = {stats['rms_reproj_error__pixels']}", file=sys.stderr)

        # Ready for a final reoptimization with the geometry
        optimization_inputs['do_optimize_extrinsics']      = True
        if not fixed_frames:
            optimization_inputs['do_optimize_frames'] = True

    else:
        optimization_inputs['lensmodel']                   = lensmodel
        if not mrcal.lensmodel_metadata_and_config(lensmodel)['has_core'] or \
           not mrcal.lensmodel_metadata_and_config(model_intrinsics.intrinsics()[0])['has_core']:
            raise Exception("I'm assuming all the models here have a core. It's just lazy coding. If you see this, feel free to fix.")

        if lensmodel == model_intrinsics.intrinsics()[0]:
            # Same model. Grab the intrinsics. They're 99% right
            optimization_inputs['intrinsics']              = intrinsics.copy()
        else:
            # Different model. Grab the intrinsics core, optimize the rest
            optimization_inputs['intrinsics']              = nps.glue(intrinsics[:,:4],
                                                                      np.zeros((Ncameras,Nintrinsics-4),),axis=-1)
        optimization_inputs['do_optimize_intrinsics_core'] = True
        optimization_inputs['do_optimize_extrinsics']      = True
        if not fixed_frames:
            optimization_inputs['do_optimize_frames'] = True

    stats = mrcal.optimize(**optimization_inputs)
    print(f"## optimized. rms = {stats['rms_reproj_error__pixels']}", file=sys.stderr)

    if not args.skip_calobject_warp_solve:
        optimization_inputs['do_optimize_calobject_warp'] = True
        stats = mrcal.optimize(**optimization_inputs)
        print(f"## optimized. rms = {stats['rms_reproj_error__pixels']}", file=sys.stderr)

    return optimization_inputs


def observation_centroid(optimization_inputs, icam):
    r'''mean pixel coordinate of all non-outlier points seen by a given camera'''

    ifcice       = optimization_inputs['indices_frame_camintrinsics_camextrinsics']
    observations = optimization_inputs['observations_board']

    # pick the camera I want
    observations = observations[ifcice[:,1] == icam]

    # ignore outliers
    q = observations[ (observations[...,2] > 0), :2]

    return np.mean(q, axis=-2)


def eval_one_rangenear_tilt(models_true,
                            range_near, range_far, tilt_radius,
                            object_width_n, object_height_n, object_spacing,
                            uncertainty_at_range_samples,
                            Ncameras,
                            Nframes_near_samples, Nframes_far_samples):

    # I want the RNG to be deterministic
    np.random.seed(0)

    uncertainties = np.zeros((len(Nframes_far_samples),
                              len(uncertainty_at_range_samples)),
                             dtype=float)


    radius_cameras = (args.camera_spacing * (Ncameras-1)) / 2.

    x_radius = args.x_radius if args.x_radius is not None else range_near*2. + radius_cameras
    y_radius = args.y_radius if args.y_radius is not None else range_near*2.
    z_radius = args.z_radius if args.z_radius is not None else range_near/10.

    # shapes (Nframes, Ncameras, Nh, Nw, 2),
    #        (Nframes, 4,3)
    q_true_near, Rt_ref_board_true_near = \
        mrcal.synthesize_board_observations(models_true,
                                            object_width_n                  = object_width_n,
                                            object_height_n                 = object_height_n,
                                            object_spacing                  = object_spacing,
                                            calobject_warp                  = calobject_warp_true_ref,
                                            rt_ref_boardcenter              = np.array((0.,  0., 0., radius_cameras, 0,  range_near,)),
                                            rt_ref_boardcenter__noiseradius = \
                                              np.array((np.pi/180. * tilt_radius,
                                                        np.pi/180. * tilt_radius,
                                                        np.pi/180. * args.roll_radius,
                                                        x_radius, y_radius, z_radius)),
                                            Nframes                         = np.max(Nframes_near_samples),
                                            which                           = args.which)
    if range_far is not None:
        q_true_far, Rt_ref_board_true_far  = \
            mrcal.synthesize_board_observations(models_true,
                                                object_width_n                  = object_width_n,
                                                object_height_n                 = object_height_n,
                                                object_spacing                  = object_spacing,
                                                calobject_warp                  = calobject_warp_true_ref,
                                                rt_ref_boardcenter              = np.array((0.,  0., 0., radius_cameras, 0,  range_far,)),
                                                rt_ref_boardcenter__noiseradius = \
                                                  np.array((np.pi/180. * tilt_radius,
                                                            np.pi/180. * tilt_radius,
                                                            np.pi/180. * args.roll_radius,
                                                            range_far*2. + radius_cameras,
                                                            range_far*2.,
                                                            range_far/10.)),
                                                Nframes                         = np.max(Nframes_far_samples),
                                                which                           = args.which)
    else:
        q_true_far            = None
        Rt_ref_board_true_far = None

    for i_Nframes_far in range(len(Nframes_far_samples)):

        Nframes_far  = Nframes_far_samples [i_Nframes_far]
        Nframes_near = Nframes_near_samples[i_Nframes_far]

        optimization_inputs = solve(Ncameras,
                                    Nframes_near, Nframes_far,
                                    object_spacing,
                                    models_true,
                                    q_true_near, Rt_ref_board_true_near,
                                    q_true_far,  Rt_ref_board_true_far)

        models_out = \
            [ mrcal.cameramodel( optimization_inputs = optimization_inputs,
                                 icam_intrinsics     = icam ) \
              for icam in range(Ncameras) ]

        model = models_out[args.icam_uncertainty]

        if args.show_geometry_first_solve:
            mrcal.show_geometry(models_out,
                                show_calobjects = True,
                                wait = True)
            sys.exit()
        if args.show_uncertainty_first_solve:
            mrcal.show_projection_uncertainty(model,
                                              method      = args.method,
                                              observations= True,
                                              wait        = True)
            sys.exit()
        if args.write_models_first_solve:
            for i in range(len(models_out)):
                f = f"/tmp/camera{i}.cameramodel"
                if os.path.exists(f):
                    input(f"File {f} already exists, and I want to overwrite it. Press enter to overwrite. Ctrl-c to exit")
                models_out[i].write(f)
                print(f"Wrote {f}")
            sys.exit()

        # shape (N,3)
        # I sample the center of the imager
        pcam_samples = \
            mrcal.unproject( observation_centroid(optimization_inputs,
                                                  args.icam_uncertainty),
                             *model.intrinsics(),
                             normalize = True) * \
                             nps.dummy(uncertainty_at_range_samples, -1)

        uncertainties[i_Nframes_far] = \
            mrcal.projection_uncertainty(pcam_samples,
                                         model,
                                         method = args.method,
                                         what   = 'worstdirection-stdev')

    return uncertainties



output_table_legend = 'range_uncertainty_sample Nframes_near Nframes_far Ncameras range_near range_far tilt_radius object_width_n object_spacing uncertainty'
output_table_fmt    = '%f %d %d %d %f %f %f %d %f %f'
output_table_icol__range_uncertainty_sample = 0
output_table_icol__Nframes_near             = 1
output_table_icol__Nframes_far              = 2
output_table_icol__Ncameras                 = 3
output_table_icol__range_near               = 4
output_table_icol__range_far                = 5
output_table_icol__tilt_radius              = 6
output_table_icol__object_width_n           = 7
output_table_icol__object_spacing           = 8
output_table_icol__uncertainty              = 9
output_table_Ncols                          = 10

output_table = np.zeros( (args.Nscan_samples, Nuncertainty_at_range_samples, output_table_Ncols), dtype=float)

output_table[:,:, output_table_icol__range_uncertainty_sample] += uncertainty_at_range_samples

if re.match("num_far_constant_Nframes_", args.scan):

    Nfar_samples = args.Nscan_samples
    if   args.scan == "num_far_constant_Nframes_near":
        Nframes_far_samples = np.linspace(0,
                                          args.Nframes_near//4,
                                          Nfar_samples, dtype=int)
        Nframes_near_samples = Nframes_far_samples*0 + args.Nframes_near

    else:
        Nframes_far_samples  = np.linspace(0,
                                           args.Nframes_all,
                                           Nfar_samples, dtype=int)
        Nframes_near_samples = args.Nframes_all - Nframes_far_samples

    models_true = \
        [ mrcal.cameramodel(intrinsics = model_intrinsics.intrinsics(),
                            imagersize = model_intrinsics.imagersize(),
                            rt_ref_cam = np.array((0,0,0,
                                                   i*args.camera_spacing,
                                                   0,0), dtype=float) ) \
          for i in range(controllable_args['Ncameras']['value']) ]

    # shape (args.Nscan_samples, Nuncertainty_at_range_samples)
    output_table[:,:, output_table_icol__uncertainty] = \
        eval_one_rangenear_tilt(models_true,
                                *controllable_args['range']['value'],
                                controllable_args['tilt_radius']['value'],
                                controllable_args['object_width_n']['value'],
                                args.object_height_n,
                                controllable_args['object_spacing']['value'],
                                uncertainty_at_range_samples,
                                controllable_args['Ncameras']['value'],
                                Nframes_near_samples, Nframes_far_samples)

    output_table[:,:, output_table_icol__Nframes_near]    += nps.transpose(Nframes_near_samples)
    output_table[:,:, output_table_icol__Nframes_far]     += nps.transpose(Nframes_far_samples)
    output_table[:,:, output_table_icol__Ncameras]        = controllable_args['Ncameras']['value']
    output_table[:,:, output_table_icol__range_near]      = controllable_args['range']['value'][0]
    output_table[:,:, output_table_icol__range_far]       = controllable_args['range']['value'][1]
    output_table[:,:, output_table_icol__tilt_radius ]    = controllable_args['tilt_radius']['value']
    output_table[:,:, output_table_icol__object_width_n ] = controllable_args['object_width_n']['value']
    output_table[:,:, output_table_icol__object_spacing ] = controllable_args['object_spacing']['value']

    samples = Nframes_far_samples

elif args.scan == "range":
    Nframes_near_samples = np.array( (controllable_args['Nframes']['value'],), dtype=int)
    Nframes_far_samples  = np.array( (0,),            dtype=int)

    models_true = \
        [ mrcal.cameramodel(intrinsics = model_intrinsics.intrinsics(),
                            imagersize = model_intrinsics.imagersize(),
                            rt_ref_cam = np.array((0,0,0,
                                                   i*args.camera_spacing,
                                                   0,0), dtype=float) ) \
          for i in range(controllable_args['Ncameras']['value']) ]

    Nrange_samples = args.Nscan_samples
    range_samples = np.linspace(*controllable_args['range']['value'],
                                Nrange_samples, dtype=float)
    for i_range in range(Nrange_samples):
        output_table[i_range,:, output_table_icol__uncertainty] = \
            eval_one_rangenear_tilt(models_true,
                                    range_samples[i_range], None,
                                    controllable_args['tilt_radius']['value'],
                                    controllable_args['object_width_n']['value'],
                                    args.object_height_n,
                                    controllable_args['object_spacing']['value'],
                                    uncertainty_at_range_samples,
                                    controllable_args['Ncameras']['value'],
                                    Nframes_near_samples, Nframes_far_samples)[0]

    output_table[:,:, output_table_icol__Nframes_near] = controllable_args['Nframes']['value']
    output_table[:,:, output_table_icol__Nframes_far]  = 0
    output_table[:,:, output_table_icol__Ncameras]     = controllable_args['Ncameras']['value']
    output_table[:,:, output_table_icol__range_near]  += nps.transpose(range_samples)
    output_table[:,:, output_table_icol__range_far]    = -1
    output_table[:,:, output_table_icol__tilt_radius ] = controllable_args['tilt_radius']['value']
    output_table[:,:, output_table_icol__object_width_n ] = controllable_args['object_width_n']['value']
    output_table[:,:, output_table_icol__object_spacing ] = controllable_args['object_spacing']['value']

    samples = range_samples

elif args.scan == "tilt_radius":
    Nframes_near_samples = np.array( (controllable_args['Nframes']['value'],), dtype=int)
    Nframes_far_samples  = np.array( (0,),            dtype=int)

    models_true = \
        [ mrcal.cameramodel(intrinsics = model_intrinsics.intrinsics(),
                            imagersize = model_intrinsics.imagersize(),
                            rt_ref_cam = np.array((0,0,0,
                                                   i*args.camera_spacing,
                                                   0,0), dtype=float) ) \
          for i in range(controllable_args['Ncameras']['value']) ]

    Ntilt_rad_samples = args.Nscan_samples
    tilt_rad_samples = np.linspace(*controllable_args['tilt_radius']['value'],
                                   Ntilt_rad_samples, dtype=float)
    for i_tilt in range(Ntilt_rad_samples):
        output_table[i_tilt,:, output_table_icol__uncertainty] = \
            eval_one_rangenear_tilt(models_true,
                                    controllable_args['range']['value'], None,
                                    tilt_rad_samples[i_tilt],
                                    controllable_args['object_width_n']['value'],
                                    args.object_height_n,
                                    controllable_args['object_spacing']['value'],
                                    uncertainty_at_range_samples,
                                    controllable_args['Ncameras']['value'],
                                    Nframes_near_samples, Nframes_far_samples)[0]

    output_table[:,:, output_table_icol__Nframes_near] = controllable_args['Nframes']['value']
    output_table[:,:, output_table_icol__Nframes_far]  = 0
    output_table[:,:, output_table_icol__Ncameras]     = controllable_args['Ncameras']['value']
    output_table[:,:, output_table_icol__range_near]   = controllable_args['range']['value']
    output_table[:,:, output_table_icol__range_far]    = -1
    output_table[:,:, output_table_icol__tilt_radius] += nps.transpose(tilt_rad_samples)
    output_table[:,:, output_table_icol__object_width_n ] = controllable_args['object_width_n']['value']
    output_table[:,:, output_table_icol__object_spacing ] = controllable_args['object_spacing']['value']

    samples = tilt_rad_samples

elif args.scan == "Ncameras":

    Nframes_near_samples = np.array( (controllable_args['Nframes']['value'],), dtype=int)
    Nframes_far_samples  = np.array( (0,),            dtype=int)

    N_Ncameras_samples = args.Nscan_samples
    Ncameras_samples = np.linspace(*controllable_args['Ncameras']['value'],
                                   N_Ncameras_samples, dtype=int)
    for i_Ncameras in range(N_Ncameras_samples):

        Ncameras = Ncameras_samples[i_Ncameras]
        models_true = \
            [ mrcal.cameramodel(intrinsics = model_intrinsics.intrinsics(),
                                imagersize = model_intrinsics.imagersize(),
                                rt_ref_cam = np.array((0,0,0,
                                                       i*args.camera_spacing,
                                                       0,0), dtype=float) ) \
              for i in range(Ncameras) ]

        output_table[i_Ncameras,:, output_table_icol__uncertainty] = \
            eval_one_rangenear_tilt(models_true,
                                    controllable_args['range']['value'], None,
                                    controllable_args['tilt_radius']['value'],
                                    controllable_args['object_width_n']['value'],
                                    args.object_height_n,
                                    controllable_args['object_spacing']['value'],
                                    uncertainty_at_range_samples,
                                    Ncameras,
                                    Nframes_near_samples, Nframes_far_samples)[0]

    output_table[:,:, output_table_icol__Nframes_near] = controllable_args['Nframes']['value']
    output_table[:,:, output_table_icol__Nframes_far]  = 0
    output_table[:,:, output_table_icol__Ncameras]    += nps.transpose(Ncameras_samples)
    output_table[:,:, output_table_icol__range_near]   = controllable_args['range']['value']
    output_table[:,:, output_table_icol__range_far]    = -1
    output_table[:,:, output_table_icol__tilt_radius]  = controllable_args['tilt_radius']['value']
    output_table[:,:, output_table_icol__object_width_n ] = controllable_args['object_width_n']['value']
    output_table[:,:, output_table_icol__object_spacing ] = controllable_args['object_spacing']['value']

    samples = Ncameras_samples

elif args.scan == "Nframes":

    Nframes_far_samples  = np.array( (0,),            dtype=int)

    models_true = \
        [ mrcal.cameramodel(intrinsics = model_intrinsics.intrinsics(),
                            imagersize = model_intrinsics.imagersize(),
                            rt_ref_cam = np.array((0,0,0,
                                                   i*args.camera_spacing,
                                                   0,0), dtype=float) ) \
          for i in range(controllable_args['Ncameras']['value']) ]

    N_Nframes_samples = args.Nscan_samples
    Nframes_samples = np.linspace(*controllable_args['Nframes']['value'],
                                  N_Nframes_samples, dtype=int)
    for i_Nframes in range(N_Nframes_samples):

        Nframes = Nframes_samples[i_Nframes]
        Nframes_near_samples = np.array( (Nframes,), dtype=int)

        output_table[i_Nframes,:, output_table_icol__uncertainty] = \
            eval_one_rangenear_tilt(models_true,
                                    controllable_args['range']['value'], None,
                                    controllable_args['tilt_radius']['value'],
                                    controllable_args['object_width_n']['value'],
                                    args.object_height_n,
                                    controllable_args['object_spacing']['value'],
                                    uncertainty_at_range_samples,
                                    controllable_args['Ncameras']['value'],
                                    Nframes_near_samples, Nframes_far_samples)[0]

    output_table[:,:, output_table_icol__Nframes_near]+= nps.transpose(Nframes_samples)
    output_table[:,:, output_table_icol__Nframes_far]  = 0
    output_table[:,:, output_table_icol__Ncameras]     = controllable_args['Ncameras']['value']
    output_table[:,:, output_table_icol__range_near]   = controllable_args['range']['value']
    output_table[:,:, output_table_icol__range_far]    = -1
    output_table[:,:, output_table_icol__tilt_radius]  = controllable_args['tilt_radius']['value']
    output_table[:,:, output_table_icol__object_width_n ] = controllable_args['object_width_n']['value']
    output_table[:,:, output_table_icol__object_spacing ] = controllable_args['object_spacing']['value']

    samples = Nframes_samples

elif args.scan == "object_width_n":

    Nframes_near_samples = np.array( (controllable_args['Nframes']['value'],), dtype=int)
    Nframes_far_samples  = np.array( (0,),            dtype=int)

    models_true = \
        [ mrcal.cameramodel(intrinsics = model_intrinsics.intrinsics(),
                            imagersize = model_intrinsics.imagersize(),
                            rt_ref_cam = np.array((0,0,0,
                                                   i*args.camera_spacing,
                                                   0,0), dtype=float) ) \
          for i in range(controllable_args['Ncameras']['value']) ]

    Nsamples = args.Nscan_samples
    samples  = np.linspace(*controllable_args['object_width_n']['value'],
                           Nsamples, dtype=int)

    # As I move the width, I adjust the spacing to keep the total board size
    # constant. The object spacing in the argument applies to the MIN value of
    # the object_width_n.
    W = (controllable_args['object_width_n']['value'][0]-1) * controllable_args['object_spacing']['value']
    object_spacing = W / samples



    for i_sample in range(Nsamples):

        output_table[i_sample,:, output_table_icol__uncertainty] = \
            eval_one_rangenear_tilt(models_true,
                                    controllable_args['range']['value'], None,
                                    controllable_args['tilt_radius']['value'],
                                    samples[i_sample],
                                    samples[i_sample],
                                    object_spacing[i_sample],
                                    uncertainty_at_range_samples,
                                    controllable_args['Ncameras']['value'],
                                    Nframes_near_samples, Nframes_far_samples)[0]

    output_table[:,:, output_table_icol__Nframes_near] = controllable_args['Nframes']['value']
    output_table[:,:, output_table_icol__Nframes_far]  = 0
    output_table[:,:, output_table_icol__Ncameras]     = controllable_args['Ncameras']['value']
    output_table[:,:, output_table_icol__range_near]   = controllable_args['range']['value']
    output_table[:,:, output_table_icol__range_far]    = -1
    output_table[:,:, output_table_icol__tilt_radius]  = controllable_args['tilt_radius']['value']
    output_table[:,:, output_table_icol__object_width_n ]+= nps.transpose(samples)
    output_table[:,:, output_table_icol__object_spacing ]+= nps.transpose(object_spacing)

elif args.scan == "object_spacing":

    Nframes_near_samples = np.array( (controllable_args['Nframes']['value'],), dtype=int)
    Nframes_far_samples  = np.array( (0,),            dtype=int)

    models_true = \
        [ mrcal.cameramodel(intrinsics = model_intrinsics.intrinsics(),
                            imagersize = model_intrinsics.imagersize(),
                            rt_ref_cam = np.array((0,0,0,
                                                   i*args.camera_spacing,
                                                   0,0), dtype=float) ) \
          for i in range(controllable_args['Ncameras']['value']) ]

    Nsamples = args.Nscan_samples
    samples  = np.linspace(*controllable_args['object_spacing']['value'],
                           Nsamples, dtype=float)

    # As I move the spacing, I leave object_width_n, letting the total board
    # size change. The object spacing in the argument applies to the MIN value
    # of the object_width_n.
    for i_sample in range(Nsamples):

        r = controllable_args['range']['value']
        if args.scan_object_spacing_compensate_range_from:
            r *= samples[i_sample]/args.scan_object_spacing_compensate_range_from

        output_table[i_sample,:, output_table_icol__uncertainty] = \
            eval_one_rangenear_tilt(models_true,
                                    r, None,
                                    controllable_args['tilt_radius']['value'],
                                    controllable_args['object_width_n']['value'],
                                    args.object_height_n,
                                    samples[i_sample],
                                    uncertainty_at_range_samples,
                                    controllable_args['Ncameras']['value'],
                                    Nframes_near_samples, Nframes_far_samples)[0]

    output_table[:,:, output_table_icol__Nframes_near] = controllable_args['Nframes']['value']
    output_table[:,:, output_table_icol__Nframes_far]  = 0
    output_table[:,:, output_table_icol__Ncameras]     = controllable_args['Ncameras']['value']
    output_table[:,:, output_table_icol__range_far]    = -1
    output_table[:,:, output_table_icol__tilt_radius]  = controllable_args['tilt_radius']['value']
    output_table[:,:, output_table_icol__object_width_n ] = controllable_args['object_width_n']['value']
    output_table[:,:, output_table_icol__object_spacing ]+= nps.transpose(samples)

    if args.scan_object_spacing_compensate_range_from:
        output_table[:,:, output_table_icol__range_near] += controllable_args['range']['value'] * nps.transpose(samples/samples[0])
    else:
        output_table[:,:, output_table_icol__range_near] = controllable_args['range']['value']

else:
    # no --scan. We just want one sample

    Nframes_near_samples = np.array( (controllable_args['Nframes']['value'],), dtype=int)
    Nframes_far_samples  = np.array( (0,),            dtype=int)

    models_true = \
        [ mrcal.cameramodel(intrinsics = model_intrinsics.intrinsics(),
                            imagersize = model_intrinsics.imagersize(),
                            rt_ref_cam = np.array((0,0,0,
                                                   i*args.camera_spacing,
                                                   0,0), dtype=float) ) \
          for i in range(controllable_args['Ncameras']['value']) ]

    output_table[0,:, output_table_icol__uncertainty] = \
        eval_one_rangenear_tilt(models_true,
                                controllable_args['range']['value'], None,
                                controllable_args['tilt_radius']['value'],
                                controllable_args['object_width_n']['value'],
                                args.object_height_n,
                                controllable_args['object_spacing']['value'],
                                uncertainty_at_range_samples,
                                controllable_args['Ncameras']['value'],
                                Nframes_near_samples, Nframes_far_samples)[0]

    output_table[:,:, output_table_icol__Nframes_near] = controllable_args['Nframes']['value']
    output_table[:,:, output_table_icol__Nframes_far]  = 0
    output_table[:,:, output_table_icol__Ncameras]     = controllable_args['Ncameras']['value']
    output_table[:,:, output_table_icol__range_near]   = controllable_args['range']['value']
    output_table[:,:, output_table_icol__range_far]    = -1
    output_table[:,:, output_table_icol__tilt_radius]  = controllable_args['tilt_radius']['value']
    output_table[:,:, output_table_icol__object_width_n ] = controllable_args['object_width_n']['value']
    output_table[:,:, output_table_icol__object_spacing ] = controllable_args['object_spacing']['value']

    samples = None



if isinstance(controllable_args['range']['value'], float):
    guides = [ f"arrow nohead dashtype 3 from {controllable_args['range']['value']},graph 0 to {controllable_args['range']['value']},graph 1" ]
else:
    guides = [ f"arrow nohead dashtype 3 from {r},graph 0 to {r},graph 1" for r in controllable_args['range']['value'] ]
guides.append(f"arrow nohead dashtype 3 from graph 0,first {args.observed_pixel_uncertainty} to graph 1,first {args.observed_pixel_uncertainty}")

title = args.title

if   args.scan == "num_far_constant_Nframes_near":
    if title is None:
        title = f"Scanning 'far' observations added to a set of 'near' observations. Have {controllable_args['Ncameras']['value']} cameras, {args.Nframes_near} 'near' observations, at ranges {controllable_args['range']['value']}."
    legend_what = 'Nframes_far'
elif args.scan == "num_far_constant_Nframes_all":
    if title is None:
        title = f"Scanning 'far' observations replacing 'near' observations. Have {controllable_args['Ncameras']['value']} cameras, {args.Nframes_all} total observations, at ranges {controllable_args['range']['value']}."
    legend_what = 'Nframes_far'
elif args.scan == "Nframes":
    if title is None:
        title = f"Scanning Nframes. Have {controllable_args['Ncameras']['value']} cameras looking out at {controllable_args['range']['value']:.2f}m."
    legend_what = 'Nframes'
elif args.scan == "Ncameras":
    if title is None:
        title = f"Scanning Ncameras. Observing {controllable_args['Nframes']['value']} boards at {controllable_args['range']['value']:.2f}m."
    legend_what = 'Ncameras'
elif args.scan == "range":
    if title is None:
        title = f"Scanning the distance to observations. Have {controllable_args['Ncameras']['value']} cameras looking at {controllable_args['Nframes']['value']} boards."
    legend_what = 'Range-to-chessboards'
elif args.scan == "tilt_radius":
    if title is None:
        title = f"Scanning the board tilt. Have {controllable_args['Ncameras']['value']} cameras looking at {controllable_args['Nframes']['value']} boards at {controllable_args['range']['value']:.2f}m"
    legend_what = 'Random chessboard tilt radius'
elif args.scan == "object_width_n":
    if title is None:
        title = f"Scanning the calibration object density, keeping the board size constant. Have {controllable_args['Ncameras']['value']} cameras looking at {controllable_args['Nframes']['value']} boards at {controllable_args['range']['value']:.2f}m"
    legend_what = 'Number of chessboard points per side'
elif args.scan == "object_spacing":
    if title is None:
        if args.scan_object_spacing_compensate_range_from:
            title = f"Scanning the calibration object spacing, keeping the point count constant, and letting the board grow. Range grows with spacing. Have {controllable_args['Ncameras']['value']} cameras looking at {controllable_args['Nframes']['value']} boards at {controllable_args['range']['value']:.2f}m"
        else:
            title = f"Scanning the calibration object spacing, keeping the point count constant, and letting the board grow. Range is constant. Have {controllable_args['Ncameras']['value']} cameras looking at {controllable_args['Nframes']['value']} boards at {controllable_args['range']['value']:.2f}m"
    legend_what = 'Distance between adjacent chessboard corners'
else:
    # no --scan. We just want one sample
    if title is None:
        title = f"Have {controllable_args['Ncameras']['value']} cameras looking at {controllable_args['Nframes']['value']} boards at {controllable_args['range']['value']:.2f}m with tilt radius {controllable_args['tilt_radius']['value']}"

if args.extratitle is not None:
    title = f"{title}: {args.extratitle}"

if samples is None:
    legend = None
elif samples.dtype.kind == 'i':
    legend = np.array([ f"{legend_what} = {x}" for x in samples])
else:
    legend = np.array([ f"{legend_what} = {x:.2f}" for x in samples])

np.savetxt(sys.stdout,
           nps.clump(output_table, n=2),
           fmt   = output_table_fmt,
           header= output_table_legend)

plotoptions = \
    dict( yrange   = (0, args.ymax),
          _with    = 'lines',
          _set     = guides,
          unset    = 'grid',
          title    = title,
          xlabel   = 'Range (m)',
          ylabel   = 'Expected worst-direction uncertainty (pixels)',
          hardcopy = args.hardcopy,
          terminal = args.terminal,
          wait     = not args.explore and args.hardcopy is None)
if legend is not None: plotoptions['legend'] = legend

if args.set:
    gp.add_plot_option(plotoptions,
                       _set = args.set)
if args.unset:
    gp.add_plot_option(plotoptions,
                       _unset = args.unset)

gp.plot(uncertainty_at_range_samples,
        output_table[:,:, output_table_icol__uncertainty],
        **plotoptions)

if args.explore:
    import IPython
    IPython.embed()
sys.exit()



#!/usr/bin/env python3

r'''Basic structure-from-motion test

I observe, with noise, a number of points from various angles with a single
camera, and I make sure that I can accurately compute the locations of the
points.

Both the cameras and observed points are free to move, so this test fixes the
positions of some points in space in order to sufficiently constrain the problem

'''

import sys
import numpy as np
import numpysane as nps
import os

testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import testutils
import test_sfm_helpers

# This many points are not optimized, and we know exactly where they are. We
# need this to set the scale of the problem. Otherwise the solve is ambiguous
Npoints_fixed = 3

add_outlier = False

############# Set up my world, and compute all the perfect positions, pixel
############# observations of everything
model,                \
imagersize,           \
lensmodel,            \
intrinsics_data,      \
indices_point_camera, \
                      \
pref_true,            \
rt_cam_ref_true,      \
qcam_true,            \
observations_true,    \
                      \
pref_noisy,           \
rt_cam_ref_noisy,     \
qcam_noisy,           \
observations_noisy =  \
    test_sfm_helpers.generate_world(Npoints_fixed)

indices_point_camintrinsics_camextrinsics = \
    nps.glue( indices_point_camera[:,(0,)],
              indices_point_camera[:,(0,)] * 0,
              indices_point_camera[:,(1,)],
              axis = -1 )


if add_outlier:
    # Make an outlier point: the first observation x coord is way off
    observations_noisy[0,0] += 100.

optimization_inputs = \
    dict( intrinsics                                = nps.atleast_dims(intrinsics_data, -2),
          rt_cam_ref                                = rt_cam_ref_noisy,
          points                                    = pref_noisy,
          observations_point                        = observations_noisy,
          indices_point_camintrinsics_camextrinsics = indices_point_camintrinsics_camextrinsics,

          lensmodel                         = lensmodel,
          imagersizes                       = nps.atleast_dims(imagersize, -2),
          Npoints_fixed                     = Npoints_fixed,
          do_optimize_intrinsics_core       = False,
          do_optimize_intrinsics_distortions= False,
          do_optimize_extrinsics            = True,
          do_optimize_frames                = True,
          do_apply_outlier_rejection        = False,
          verbose                           = False)

stats = mrcal.optimize(**optimization_inputs)


# If we have outliers, I need to be able to detect it. If I can detect it,
# markOutliers() should do that, mark it, and re-optimize without that
# observation. This code (disabled currently) shows that we can't detect it for
# discrete points, so markOutliers() ignores discrete points. If you want to
# figure out how to do this right, start by re-enabling this code
if add_outlier:
    p,x,J,f = mrcal.optimizer_callback(**optimization_inputs)

    if 0: # print measurements x

        import gnuplotlib as gp
        gp.plot(x, _with='points pt 7', wait=1)
        sys.exit()

    elif 1: # print outlierness. Not 100% sure this code is right
        J = np.array(J.todense())

        outlierness = np.zeros((len(observations_noisy),), dtype=float)
        for iobservation in range(len(observations_noisy)):

            Ji = J[iobservation*3:iobservation*3+3, :]
            xi = x[iobservation*3:iobservation*3+3   ]

            # A = J* inv(JtJ) J*t
            # B = inv(A - I)
            # Dima's self+others:        x*t (-B      ) x*

            A = \
                nps.matmult( f.solve_xt_JtJ_bt(Ji),
                             nps.transpose(Ji) )
            B = np.linalg.inv(A - np.eye(3))

            outlierness[iobservation] = -nps.inner(xi, nps.inner(B,xi))


        import gnuplotlib as gp
        gp.plot(outlierness, _with='points pt 7', wait=1)
        sys.exit()




testutils.confirm_equal(pref_noisy,
                        pref_true,
                        msg = f"Solved at ref coords with known-position points",
                        eps = 0.1)

testutils.finish()



#!/usr/bin/env python3

r'''Basic camera-calibration test

I observe, with noise, a number of chessboards from various angles with several
cameras. And I make sure that I can more or less compute the camera intrinsics
and extrinsics

'''

import sys
import numpy as np
import numpysane as nps
import os

testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import testutils

from test_calibration_helpers import sample_dqref
import copy

# I want the RNG to be deterministic
np.random.seed(0)

############# Set up my world, and compute all the perfect positions, pixel
############# observations of everything
models_ref = ( mrcal.cameramodel(f"{testdir}/data/cam0.opencv8.cameramodel"),
               mrcal.cameramodel(f"{testdir}/data/cam0.opencv8.cameramodel"),
               mrcal.cameramodel(f"{testdir}/data/cam1.opencv8.cameramodel"),
               mrcal.cameramodel(f"{testdir}/data/cam1.opencv8.cameramodel") )

imagersizes = nps.cat( *[m.imagersize() for m in models_ref] )
lensmodel   = models_ref[0].intrinsics()[0]
# I have opencv8 models_ref, but let me truncate to opencv4 models_ref to keep this
# simple and fast
lensmodel = 'LENSMODEL_OPENCV4'
for m in models_ref:
    m.intrinsics( intrinsics = (lensmodel, m.intrinsics()[1][:8]))
Nintrinsics = mrcal.lensmodel_num_params(lensmodel)

Ncameras = len(models_ref)
Nframes  = 50

models_ref[0].rt_cam_ref(np.zeros((6,), dtype=float))
models_ref[1].rt_cam_ref(np.array((0.08,0.2,0.02, 1., 0.9,0.1)))
models_ref[2].rt_cam_ref(np.array((0.01,0.07,0.2, 2.1,0.4,0.2)))
models_ref[3].rt_cam_ref(np.array((-0.1,0.08,0.08, 4.4,0.2,0.1)))


pixel_uncertainty_stdev = 1.5
object_spacing          = 0.1
object_width_n          = 10
object_height_n         = 9
calobject_warp_ref      = np.array((0.002, -0.005))

# shapes (Nframes, Ncameras, Nh, Nw, 2),
#        (Nframes, 4,3)
q_ref,Rt_ref_board_ref = \
    mrcal.synthesize_board_observations(models_ref,
                                        object_width_n                  = object_width_n,
                                        object_height_n                 = object_height_n,
                                        object_spacing                  = object_spacing,
                                        calobject_warp                  = calobject_warp_ref,
                                        rt_ref_boardcenter              = np.array((0.,  0.,  0., -2,   0,  4.0)),
                                        rt_ref_boardcenter__noiseradius = np.array((np.pi/180.*30., np.pi/180.*30., np.pi/180.*20., 2.5, 2.5, 2.0)),
                                        Nframes                         = Nframes)
frames_ref = mrcal.rt_from_Rt(Rt_ref_board_ref)

############# I have perfect observations in q_ref. I corrupt them by noise
# weight has shape (Nframes, Ncameras, Nh, Nw),
weight01 = (np.random.rand(*q_ref.shape[:-1]) + 1.) / 2. # in [0,1]
weight0 = 0.2
weight1 = 1.0
weight = weight0 + (weight1-weight0)*weight01

# I want observations of shape (Nframes*Ncameras, Nh, Nw, 3) where each row is
# (x,y,weight)
observations_ref = nps.clump( nps.glue(q_ref,
                                       nps.dummy(weight,-1),
                                       axis=-1),
                              n=2)

q_noise,observations = sample_dqref(observations_ref,
                                    pixel_uncertainty_stdev,
                                    make_outliers = True)

# Now I make some of the observations bogus, and mark them as input outliers.
# The solve should be robust to that, but any code that uses the bogus data
# DESPITE it being marked as bogus will generate a test failure
#
# Let's pretend the center of the chessboard has an apriltag, so all those
# observations are bogus. I block out a 5x5 chunk in the center
i0 = object_height_n//2
j0 = object_width_n//2
observations[..., i0-2:i0+3,j0-2:j0+3, 2] = -1.    # weight<=0: outlier
observations[..., i0-2:i0+3,j0-2:j0+3,:2] = -100.0 # all the values are bogus

############# Now I pretend that the noisy observations are all I got, and I run
############# a calibration from those

# Dense observations. All the cameras see all the boards
indices_frame_camera = np.zeros( (Nframes*Ncameras, 2), dtype=np.int32)
indices_frame = indices_frame_camera[:,0].reshape(Nframes,Ncameras)
indices_frame.setfield(nps.outer(np.arange(Nframes, dtype=np.int32),
                                 np.ones((Ncameras,), dtype=np.int32)),
                       dtype = np.int32)
indices_camera = indices_frame_camera[:,1].reshape(Nframes,Ncameras)
indices_camera.setfield(nps.outer(np.ones((Nframes,), dtype=np.int32),
                                 np.arange(Ncameras, dtype=np.int32)),
                       dtype = np.int32)

indices_frame_camintrinsics_camextrinsics = \
    nps.glue(indices_frame_camera,
             indices_frame_camera[:,(1,)]-1,
             axis=-1)


intrinsics_data,rt_cam_ref,rt_ref_frame = \
    mrcal.seed_stereographic(imagersizes          = imagersizes,
                             focal_estimate       = 1500,
                             indices_frame_camera = indices_frame_camera,
                             observations         = observations,
                             object_spacing       = object_spacing)

# I have a stereographic intrinsics estimate. Mount it into a full distortiony
# model, seeded with random numbers
intrinsics = np.zeros((Ncameras,Nintrinsics), dtype=float)
intrinsics[:,:4] = intrinsics_data
intrinsics[:,4:] = np.random.random( (Ncameras, intrinsics.shape[1]-4) ) * 1e-6

optimization_inputs = \
    dict( intrinsics                                = intrinsics,
          rt_cam_ref                                = rt_cam_ref,
          rt_ref_frame                              = rt_ref_frame,
          points                                    = None,
          observations_board                        = observations,
          indices_frame_camintrinsics_camextrinsics = indices_frame_camintrinsics_camextrinsics,
          observations_point                        = None,
          indices_point_camintrinsics_camextrinsics = None,
          lensmodel                                 = lensmodel,
          calobject_warp                            = None,
          imagersizes                               = imagersizes,
          calibration_object_spacing                = object_spacing,
          verbose                                   = False,
          do_apply_regularization                   = True)

# Solve this thing incrementally
optimization_inputs['do_optimize_intrinsics_core']        = False
optimization_inputs['do_optimize_intrinsics_distortions'] = False
optimization_inputs['do_optimize_extrinsics']             = True
optimization_inputs['do_optimize_frames']                 = True
optimization_inputs['do_optimize_calobject_warp']         = False
mrcal.optimize(**optimization_inputs,
               do_apply_outlier_rejection = True)

optimization_inputs['do_optimize_intrinsics_core']        = True
optimization_inputs['do_optimize_intrinsics_distortions'] = False
optimization_inputs['do_optimize_extrinsics']             = True
optimization_inputs['do_optimize_frames']                 = True
optimization_inputs['do_optimize_calobject_warp']         = False
mrcal.optimize(**optimization_inputs,
               do_apply_outlier_rejection = True)

testutils.confirm_equal( mrcal.num_states(**optimization_inputs),
                         4*Ncameras + 6*(Ncameras-1) + 6*Nframes,
                         msg="num_states()")
testutils.confirm_equal( mrcal.num_states_intrinsics(**optimization_inputs),
                         4*Ncameras,
                         msg="num_states_intrinsics()")
testutils.confirm_equal( mrcal.num_intrinsics_optimization_params(**optimization_inputs),
                         4,
                         msg="num_intrinsics_optimization_params()")
testutils.confirm_equal( mrcal.num_states_extrinsics(**optimization_inputs),
                         6*(Ncameras-1),
                         msg="num_states_extrinsics()")
testutils.confirm_equal( mrcal.num_states_frames(**optimization_inputs),
                         6*Nframes,
                         msg="num_states_frames()")
testutils.confirm_equal( mrcal.num_states_points(**optimization_inputs),
                         0,
                         msg="num_states_points()")
testutils.confirm_equal( mrcal.num_states_calobject_warp(**optimization_inputs),
                         0,
                         msg="num_states_calobject_warp()")

testutils.confirm_equal( mrcal.num_measurements_boards(**optimization_inputs),
                         object_width_n*object_height_n*2*Nframes*Ncameras,
                         msg="num_measurements_boards()")
testutils.confirm_equal( mrcal.num_measurements_points(**optimization_inputs),
                         0,
                         msg="num_measurements_points()")
testutils.confirm_equal( mrcal.num_measurements_regularization(**optimization_inputs),
                         Ncameras * 2,
                         msg="num_measurements_regularization()")


optimization_inputs['do_optimize_intrinsics_core']        = True
optimization_inputs['do_optimize_intrinsics_distortions'] = True
optimization_inputs['do_optimize_extrinsics']             = True
optimization_inputs['do_optimize_frames']                 = True
optimization_inputs['do_optimize_calobject_warp']         = True

optimization_inputs['calobject_warp'] = np.array((0.001, 0.001))
stats = mrcal.optimize(**optimization_inputs,
                       do_apply_outlier_rejection = True)

rmserr = stats['rms_reproj_error__pixels']


testutils.confirm_equal( mrcal.state_index_intrinsics(2, **optimization_inputs),
                         8*2,
                         msg="state_index_intrinsics()")
testutils.confirm_equal( mrcal.state_index_extrinsics(2, **optimization_inputs),
                         8*Ncameras + 6*2,
                         msg="state_index_extrinsics()")
testutils.confirm_equal( mrcal.state_index_frames(2, **optimization_inputs),
                         8*Ncameras + 6*(Ncameras-1) + 6*2,
                         msg="state_index_frames()")
testutils.confirm_equal( mrcal.state_index_calobject_warp(**optimization_inputs),
                         8*Ncameras + 6*(Ncameras-1) + 6*Nframes,
                         msg="state_index_calobject_warp()")

testutils.confirm_equal( mrcal.measurement_index_boards(2, **optimization_inputs),
                         object_width_n*object_height_n*2* 2,
                         msg="measurement_index_boards()")
testutils.confirm_equal( mrcal.measurement_index_regularization(**optimization_inputs),
                         object_width_n*object_height_n*2*Nframes*Ncameras,
                         msg="measurement_index_regularization()")


############# Calibration computed. Now I see how well I did
models_solved = \
    [ mrcal.cameramodel( optimization_inputs = optimization_inputs,
                         icam_intrinsics     = i )
      for i in range(Ncameras)]

if False:
    for i in range(0,Ncameras):
        f = f'/tmp/tst{i}.cameramodel'
        models_solved[i].write(f)
        print(f"Wrote '{f}'")

testutils.confirm_equal(rmserr, 0,
                        eps = 2.5,
                        msg = "Converged to a low RMS error")

testutils.confirm_equal( optimization_inputs['calobject_warp'],
                         calobject_warp_ref,
                         eps = 2e-3,
                         msg = "Recovered the calibration object shape" )

testutils.confirm_equal( np.std( mrcal.measurements_board(optimization_inputs,
                                                          x = stats['x'])),
                         pixel_uncertainty_stdev,
                         eps = pixel_uncertainty_stdev*0.1,
                         msg = "Residual have the expected distribution" )

# Checking the extrinsics. These aren't defined absolutely: each solve is free
# to put the observed frames anywhere it likes. The projection-diff code
# computes a transformation to address this. Here I simply look at the relative
# transformations between cameras, which would cancel out any extra
# transformations, AND since camera0 is fixed at the identity transformation, I
# can simply look at each extrinsics transformation.
for icam in range(1,len(models_ref)):

    Rt_extrinsics_err = \
        mrcal.compose_Rt( models_solved[icam].Rt_cam_ref(),
                          models_ref   [icam].Rt_ref_cam() )

    testutils.confirm_equal( nps.mag(Rt_extrinsics_err[3,:]),
                             0.0,
                             eps = 0.05,
                             msg = f"Recovered extrinsic translation for camera {icam}")

    testutils.confirm_equal( (np.trace(Rt_extrinsics_err[:3,:]) - 1) / 2.,
                             1.0,
                             eps = np.cos(1. * np.pi/180.0), # 1 deg
                             msg = f"Recovered extrinsic rotation for camera {icam}")

Rt_frame_err = \
    mrcal.compose_Rt( mrcal.Rt_from_rt(optimization_inputs['rt_ref_frame']),
                      mrcal.invert_Rt(Rt_ref_board_ref) )

testutils.confirm_equal( np.max(nps.mag(Rt_frame_err[..., 3,:])),
                         0.0,
                         eps = 0.08,
                         msg = "Recovered frame translation")
testutils.confirm_equal( np.min( (nps.trace(Rt_frame_err[..., :3,:]) - 1)/2. ),
                         1.0,
                         eps = np.cos(1. * np.pi/180.0), # 1 deg
                         msg = "Recovered frame rotation")


# Checking the intrinsics. Each intrinsics vector encodes an implicit
# transformation. I compute and apply this transformation when making my
# intrinsics comparisons. I make sure that within some distance of the pixel
# center, the projections match up to within some number of pixels
Nw = 60
def projection_diff(models_ref, max_dist_from_center):
    lensmodels      = [model.intrinsics()[0] for model in models_ref]
    intrinsics_data = [model.intrinsics()[1] for model in models_ref]

    # v  shape (...,Ncameras,Nheight,Nwidth,...)
    # q0 shape (...,         Nheight,Nwidth,...)
    v,q0 = \
        mrcal.sample_imager_unproject(Nw,None,
                                      *imagersizes[0],
                                      lensmodels, intrinsics_data,
                                      normalize = True)

    W,H = imagersizes[0]
    focus_center = None
    focus_radius = -1
    if focus_center is None: focus_center = ((W-1.)/2., (H-1.)/2.)
    if focus_radius < 0:     focus_radius = min(W,H)/6.


    implied_Rt10 = \
        mrcal.implied_Rt10__from_unprojections(q0,
                                               v[0,...], v[1,...],
                                               focus_center = focus_center,
                                               focus_radius = focus_radius)

    q1 = mrcal.project( mrcal.transform_point_Rt(implied_Rt10,
                                                 v[0,...]),
                       lensmodels[1], intrinsics_data[1])
    diff = nps.mag(q1 - q0)

    # zero-out everything too far from the center
    center = (imagersizes[0] - 1.) / 2.
    diff[ nps.norm2(q0 - center) > max_dist_from_center*max_dist_from_center ] = 0
    # gp.plot(diff,
    #         ascii = True,
    #         using = mrcal.imagergrid_using(imagersizes[0], Nw),
    #         square=1, _with='image', tuplesize=3, hardcopy='/tmp/yes.gp', cbmax=3)

    return diff


for icam in range(len(models_ref)):
    diff = projection_diff( (models_ref[icam], models_solved[icam]), 800)

    testutils.confirm_equal(diff, 0,
                            worstcase = True,
                            eps = 6.,
                            msg = f"Recovered intrinsics for camera {icam}")


# It would be nice to check the outlier detections, but this is iffy. Here I'm
# generating 1% outliers (hard-coded in sample_dqref()), but the outlier
# rejection is overly aggressive. I'm currently seeing 4.4%:
#
#   np.count_nonzero(observations[...,2]<=0) / observations[...,0].ravel().size
#
# The outlier rejection scheme just cuts out 3sigma residuals and above, so it's
# not great. I'm not entirely sure why it's over-reporting the outliers here,
# but I should investigate that at the same time as I overhaul the outlier
# rejection scheme (presumably to use one of my flavors of Cook's D factor)

# I test make_perfect_observations(). Doing it here is easy; doing it elsewhere
# it much more work
if True:
    optimization_inputs_perfect = copy.deepcopy(optimization_inputs)

    mrcal.make_perfect_observations(optimization_inputs_perfect,
                                    observed_pixel_uncertainty=0)
    x = mrcal.optimizer_callback(**optimization_inputs_perfect,
                                 no_jacobian      = True,
                                 no_factorization = True)[1]

    Nmeas = mrcal.num_measurements_boards(**optimization_inputs_perfect)
    if Nmeas > 0:
        i_meas0 = mrcal.measurement_index_boards(0, **optimization_inputs_perfect)
        testutils.confirm_equal( x[i_meas0:i_meas0+Nmeas],
                                 0,
                                 worstcase = True,
                                 eps = 1e-8,
                                 msg = f"make_perfect_observations() works for boards")
    else:
        testutils.confirm( False,
                           msg = f"Nmeasurements_boards <= 0")
testutils.finish()



#!/usr/bin/python3

# Copyright (c) 2017-2023 California Institute of Technology ("Caltech"). U.S.
# Government sponsorship acknowledged. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0

r'''Read/write camera models using the legacy .cahvor file format

The .cahvor functionality is available via the mrcal.cameramodel class. There's
no reason for end users to use the mrcal.cahvor module directly.

mrcal supports the .cahvor file format to interoperate with tools that work with
that format only.

Models stored as .cahvor support a subset of .cameramodel functionality: the
optimization_inputs are not included, so uncertainty computations are not
possible from .cahvor models.

CAHVOR and CAHVORE lens models can be stored in a .cahvor file as one would
expect. OPENCV models write a CAHV model, with distortions in a magic comment,
and it is up to the parser to interpret that comment. Other lens model types are
not supported.

Unless you're interfacing with tools that expect .cahvor files, there's no
reason to use this module.

'''


import sys
import re

import numpy     as np
import numpysane as nps

import mrcal

def _HVs_HVc_HVp(A,H,V):
    r'''Given a cahvor dict returns a tuple containing (Hs,Vs,Hc,Vc,Hp,Vp)'''

    Hc   = nps.inner(H, A)
    hshp = H - Hc * A
    Hs   = np.sqrt(nps.inner(hshp,hshp))

    Vc   = nps.inner(V, A)
    vsvp = V - Vc * A
    Vs   = np.sqrt(nps.inner(vsvp,vsvp))

    Hp   = hshp / Hs
    Vp   = vsvp / Vs

    return Hs,Vs,Hc,Vc,Hp,Vp

def _construct_model(C,A,H,V,
                     O = None,
                     R = None,
                     E = None,
                     *,
                     cahvore_linearity,
                     is_cahvor_or_cahvore,
                     is_cahvore,
                     Dimensions,
                     name,
                     VALID_INTRINSICS_REGION = None,
                     lensmodel_fallback   = None,
                     distortions_fallback = None,
                     **rest):
    r'''Construct a mrcal.cameramodel object from cahvor chunks

I'm going to be calling this from the outside, but everything about cahvor is a
massive hack, so I'm not documenting this
    '''

    # normalize optical axis. Mostly this is here to smooth out ASCII roundoff
    # errors
    A /= nps.mag(A)
    fx,fy,cx,cy,Hp,Vp = _HVs_HVc_HVp(A,H,V)

    # By construction Hp and Vp will both be orthogonal to A. But CAHVOR allows
    # non-orthogonal Hp,Vp. MY implementation does not support this, so I check,
    # and barf if I encounter non-orthogonal Hp,Vp
    Vp_expected = np.cross(A, Hp)
    th = np.arccos(np.clip( nps.inner(Vp,Vp_expected),
                            -1, 1)) *180./np.pi
    if th > 1e-3:
        print(f"WARNING: parsed .cahvor file has non-orthogonal Hp,Vp. Skew of {th:.3f} degrees. I'm using an orthogonal Vp, so the resulting model will work slightly differently",
              file=sys.stderr)
    Vp = Vp_expected

    R_toref = nps.transpose( nps.cat( Hp,
                                      Vp,
                                      A ))
    t_toref = C

    lensmodel   = lensmodel_fallback
    distortions = distortions_fallback
    if is_cahvor_or_cahvore:
        if O is None:
            alpha = 0
            beta  = 0
        else:
            o     = nps.matmult( O, R_toref )
            alpha = np.arctan2(o[0], o[2])
            beta  = np.arcsin( o[1] )

        if is_cahvore:
            # CAHVORE
            if E is None:
                raise Exception('Cahvor file {} LOOKS like a cahvore, but lacks the E'.format(name))
            R0,R1,R2 = R.ravel()
            E0,E1,E2 = E.ravel()

            distortions      = np.array((alpha,beta,R0,R1,R2,E0,E1,E2), dtype=float)
            lensmodel = f'LENSMODEL_CAHVORE_linearity={cahvore_linearity}'

        else:
            # CAHVOR
            if E is not None:
                raise Exception('Cahvor file {} LOOKS like a cahvor, but has an E'.format(name))

            if R is None or np.linalg.norm(R) < 1e-8:
                # pinhole
                distortions = np.array(())
                lensmodel = 'LENSMODEL_PINHOLE'
            else:
                R0,R1,R2 = R.ravel()
                distortions = np.array((alpha,beta,R0,R1,R2), dtype=float)
                lensmodel = 'LENSMODEL_CAHVOR'

    return mrcal.cameramodel(imagersize = Dimensions[:2].astype(np.int32),
                             intrinsics = (lensmodel, nps.glue( np.array((fx,fy,cx,cy), dtype=float),
                                                                distortions,
                                                                axis = -1)),
                             valid_intrinsics_region = VALID_INTRINSICS_REGION,
                             Rt_ref_cam = np.ascontiguousarray(nps.glue(R_toref,t_toref, axis=-2)))

def _read(s, name):
    r'''Reads a .cahvor file into a cameramodel

    The input is the .cahvor file contents as a string'''


    re_f = r'[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?'
    re_u = r'\d+'
    re_d = r'[-+]?\d+'
    re_s = r'.+'

    # I parse all key=value lines into my dict as raw text. Further I
    # post-process some of these raw lines.
    x = {}
    for l in s.splitlines():
        if re.match(r'^\s*#|^\s*$', l):
            continue

        m = re.match(r'\s*(\w+)\s*=\s*(.+?)\s*\n?$',
                     l, flags=re.I)
        if m:
            key   = m.group(1)
            value = m.group(2)

            # for compatibility
            if re.match('^DISTORTION', key):
                key = key.replace('DISTORTION', 'LENSMODEL')

            x[key] = value


    # Done reading. Any values that look like numbers, I convert to numbers.
    for i in x:
        if re.match('{}$'.format(re_f), x[i]):
            x[i] = float(x[i])

    # I parse the fields I know I care about into numpy arrays
    for i in ('Dimensions','C','A','H','V','O','R','E',
              'LENSMODEL_OPENCV4',
              'LENSMODEL_OPENCV5',
              'LENSMODEL_OPENCV8',
              'LENSMODEL_OPENCV12',
              'VALID_INTRINSICS_REGION'):
        if i in x:
            # Any data that's composed only of digits and whitespaces (no "."),
            # use integers
            if re.match(r'[0-9\s]+$', x[i]): totype = int
            else:                            totype = float
            x[i] = np.array( [ totype(v) for v in re.split(r'\s+', x[i])], dtype=totype)

    # Now I sanity-check the results and call it done
    for k in ('Dimensions','C','A','H','V'):
        if not k in x:
            raise Exception("Cahvor file '{}' incomplete. Missing values for: {}".
                            format(name, k))


    is_cahvore           = False
    cahvore_linearity    = None
    is_cahvor_or_cahvore = False

    if 'LENSMODEL_OPENCV12' in x:
        distortions = x["LENSMODEL_OPENCV12"]
        lensmodel = 'LENSMODEL_OPENCV12'
    elif 'LENSMODEL_OPENCV8' in x:
        distortions = x["LENSMODEL_OPENCV8"]
        lensmodel = 'LENSMODEL_OPENCV8'
    elif 'LENSMODEL_OPENCV5' in x:
        distortions = x["LENSMODEL_OPENCV5"]
        lensmodel = 'LENSMODEL_OPENCV5'
    elif 'LENSMODEL_OPENCV4' in x:
        distortions = x["LENSMODEL_OPENCV4"]
        lensmodel = 'LENSMODEL_OPENCV4'
    elif 'R' not              in x:
        distortions = np.array(())
        lensmodel = 'LENSMODEL_PINHOLE'
    else:
        is_cahvor_or_cahvore = True
        lensmodel   = None
        distortions = None

    if 'VALID_INTRINSICS_REGION' in x:
        x['VALID_INTRINSICS_REGION'] = \
            x['VALID_INTRINSICS_REGION'].reshape( len(x['VALID_INTRINSICS_REGION'])//2, 2)

    # get extrinsics from cahvor
    if 'Model' not in x:
        x['Model'] = ''

    # One of these:
    #   CAHVORE1
    #   CAHVORE2
    #   CAHVORE3,0.44
    m = re.match(r'CAHVORE\s*([0-9]+)(\s*,\s*([0-9\.e-]+))?',x['Model'])
    if m:
        modelname = x['Model']
        is_cahvore = True
        try:
            mtype = int(m.group(1))
            if m.group(3) is None:
                cahvore_linearity = None
            else:
                cahvore_linearity = float(m.group(3))
        except:
            raise Exception(f"Cahvor file '{name}' looks like CAHVORE, but the CAHVORE declaration is unparseable: '{modelname}'")

        if mtype == 1:
            if cahvore_linearity is not None:
                if cahvore_linearity != 1:
                    raise Exception(f"Cahvor file '{name}' looks like CAHVORE, but has an unexpected linearity defined. mtype=1 so I expected no linearity at all or linearity=1, but got {cahvore_linearity}. CAHVORE declaration: '{modelname}'")
            cahvore_linearity = 1
        elif mtype == 2:
            if cahvore_linearity is not None:
                if cahvore_linearity != 0:
                    raise Exception(f"Cahvor file '{name}' looks like CAHVORE, but has an unexpected linearity defined. mtype=2 so I expected no linearity at all or linearity=0, but got {cahvore_linearity}. CAHVORE declaration: '{modelname}'")
            cahvore_linearity = 0
        elif mtype == 3:
            if cahvore_linearity is None:
                raise Exception(f"Cahvor file '{name}' looks like CAHVORE, but has a missing linearity. mtype=3 a linearity parameter MUST be defined. CAHVORE declaration: '{modelname}'")
        else:
            raise Exception(f"Cahvor file '{name}' looks like CAHVORE, but has mtype={mtype}. I only know about types 1,2,3. CAHVORE declaration: '{modelname}'")
    else:
        is_cahvore = False

    return _construct_model(**x,
                            is_cahvor_or_cahvore = is_cahvor_or_cahvore,
                            is_cahvore           = is_cahvore,
                            cahvore_linearity    = cahvore_linearity,
                            name                 = name,
                            distortions_fallback = distortions,
                            lensmodel_fallback   = lensmodel)


def read(f):
    r'''Reads a .cahvor file into a cameramodel

    The input is a filename or an opened file'''

    if type(f) is mrcal.cameramodel:
        return f

    if type(f) is str:
        with open(f, 'r') as openedfile:
            return _read(openedfile.read(), f)

    return _read(f.read(), f.name)

def read_from_string(s):
    return _read(s, "<string>")

def _deconstruct_model(model):

    x = dict()

    lensmodel,intrinsics = model.intrinsics()
    m = re.match(r'^LENSMODEL_CAHVORE_linearity=([0-9\.]+)$', lensmodel)
    if m is not None:
        x['cahvore_linearity'] = float(m.group(1))
    else:
        x['cahvore_linearity'] = None

    x['Dimensions'] = model.imagersize()


    fx,fy,cx,cy = intrinsics[:4]
    Rt_toref = model.Rt_ref_cam()
    R_toref = Rt_toref[:3,:]
    t_toref = Rt_toref[ 3,:]

    x['C'] = t_toref
    x['A'] = R_toref[:,2]
    Hp     = R_toref[:,0]
    Vp     = R_toref[:,1]
    x['H'] = fx*Hp + x['A']*cx
    x['V'] = fy*Vp + x['A']*cy

    x['O'] = None
    x['R'] = None
    x['E'] = None

    if re.match('LENSMODEL_CAHVOR', lensmodel):
        # CAHVOR(E)
        alpha,beta,R0,R1,R2 = intrinsics[4:9]

        s_al,c_al,s_be,c_be = np.sin(alpha),np.cos(alpha),np.sin(beta),np.cos(beta)
        x['O'] = nps.matmult( R_toref, nps.transpose(np.array(( s_al*c_be, s_be, c_al*c_be ), dtype=float)) ).ravel()
        x['R'] = np.array((R0, R1, R2), dtype=float)

        if re.match('LENSMODEL_CAHVORE', lensmodel):
            x['E'] = intrinsics[9:]

    x['VALID_INTRINSICS_REGION'] = model.valid_intrinsics_region()

    return x


def _write(f, m, note=None):
    r'''Writes a cameramodel as a .cahvor to a writeable file object'''

    x = _deconstruct_model(m)


    if note is not None:
        for l in note.splitlines():
            f.write('# ' + l + '\n')

    lensmodel,intrinsics = m.intrinsics()
    if lensmodel == 'LENSMODEL_CAHVOR':
        f.write("Model = CAHVOR = perspective, distortion\n")
    elif re.match('LENSMODEL_(OPENCV.*|PINHOLE)', lensmodel):
        f.write("Model = CAHV = perspective, linear\n")
    else:
        if x['cahvore_linearity'] is not None:
            f.write(f"Model = CAHVORE3,{x['cahvore_linearity']} = general\n")
        else:
            raise Exception(f"Don't know how to handle lens model '{lensmodel}'")

    f.write('Dimensions = {} {}\n'.format(int(x['Dimensions'][0]), int(x['Dimensions'][1])))

    f.write(("{} =" + (" {:15.10f}" * 3) + "\n").format('C', *x['C']))
    f.write(("{} =" + (" {:15.10f}" * 3) + "\n").format('A', *x['A']))
    f.write(("{} =" + (" {:15.10f}" * 3) + "\n").format('H', *x['H']))
    f.write(("{} =" + (" {:15.10f}" * 3) + "\n").format('V', *x['V']))

    if re.match('LENSMODEL_CAHVOR', lensmodel):
        # CAHVOR(E)
        f.write(("{} =" + (" {:15.10f}" * 3) + "\n").format('O', *x['O']))
        f.write(("{} =" + (" {:15.10f}" * 3) + "\n").format('R', *x['R']))

        if re.match('LENSMODEL_CAHVORE', lensmodel):
            f.write(("{} =" + (" {:15.10f}" * 3) + "\n").format('E', *x['E']))

    elif re.match('LENSMODEL_OPENCV', lensmodel):
        Ndistortions = mrcal.lensmodel_num_params(lensmodel) - 4
        f.write(("{} =" + (" {:15.10f}" * Ndistortions) + "\n").format(lensmodel, *intrinsics[4:]))
    elif lensmodel == 'LENSMODEL_PINHOLE':
        # the CAHV values we already wrote are all that's needed
        pass
    else:
        raise Exception(f"Cannot write lens model '{lensmodel}' to a .cahvor file. I only support PINHOLE, CAHVOR(E) and OPENCV model")

    if x['VALID_INTRINSICS_REGION'] is not None:
        f.write("VALID_INTRINSICS_REGION = ")
        np.savetxt(f, x['VALID_INTRINSICS_REGION'].ravel(), fmt='%.2f', newline=' ')
        f.write('\n')

    # Write covariance matrix. Old jplv parser requires that this exists, even
    # if the actual values don't matter
    S_size = 12
    if   re.match('LENSMODEL_CAHVORE', lensmodel): S_size = 21
    elif re.match('LENSMODEL_CAHVOR',  lensmodel): S_size = 18
    f.write("S =\n" + ((" 0.0" * S_size) + "\n") * S_size)

    # Extra spaces before "=" are significant. Old jplv parser gets confused if
    # they don't exist
    Hs,Vs,Hc,Vc = intrinsics[:4]
    f.write("Hs    = {}\n".format(Hs))
    f.write("Hc    = {}\n".format(Hc))
    f.write("Vs    = {}\n".format(Vs))
    f.write("Vc    = {}\n".format(Vc))
    f.write("# this is hard-coded\nTheta = {} (-90.0 deg)\n".format(-np.pi/2))

    # Write internal covariance matrix. Again, the old jplv parser requires that
    # this exists, even if the actual values don't matter
    S_internal_size = 5
    f.write("S internal =\n" + ((" 0.0" * S_internal_size) + "\n") * S_internal_size)

    return True

def write(f, m, note=None):
    r'''Writes a cameramodel as a .cahvor to a filename or a writeable file object'''

    if type(f) is str:
        with open(f, 'w') as openedfile:
            return _write(openedfile, m, note)

    return _write(f, m)

def read_transforms(f):
    r'''Reads a file (a filename string, or a file-like object: an iterable
    containing lines of text) into a transforms dict, and returns the dict

    '''

    needclose = False
    if type(f) is str:
        filename = f
        f = open(filename, 'r')
        needclose = True

    x = { 'veh_from_ins': None,

          # this is actually "pair" to ins
          'ins_from_camera': {} }

    for l in f:
        if re.match(r'^\s*#|^\s*$', l):
            continue

        re_f = r'[-+]?(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][-+]?\d+)?'
        re_u = r'\d+'
        re_d = r'[-+]?\d+'
        re_s = r'.+'

        re_pos  = r'\(\s*({f})\s+({f})\s+({f})\s*\)'        .format(f=re_f)
        re_quat = r'\(\s*({f})\s+({f})\s+({f})\s+({f})\s*\)'.format(f=re_f)
        m = re.match(r'\s*ins2veh\s*=\s*{p}\s*{q}\s*\n?$'.
                     format(u=re_u, p=re_pos, q=re_quat),
                     l)
        if m:
            if x['veh_from_ins'] is not None:
                raise("'{}' is corrupt: more than one 'ins2veh'".format(f.name))

            x['veh_from_ins'] = mrcal.Rt_from_qt( np.array((float(m.group(4)),float(m.group(5)),float(m.group(6)),float(m.group(7)),
                                                            float(m.group(1)),float(m.group(2)),float(m.group(3)),
                                                            )))
            continue

        m = re.match(r'\s*cam2ins\s*\[({u})\]\s*=\s*{p}\s*{q}\s*\n?$'.
                     format(u=re_u, p=re_pos, q=re_quat),
                     l)
        if m:
            i = int(m.group(1))
            if x['ins_from_camera'].get(i) is not None:
                raise("'{}' is corrupt: more than one 'cam2ins'[{}]".format(f.name, i))

            x['ins_from_camera'][i] = mrcal.Rt_from_qt( np.array((float(m.group(5)),float(m.group(6)),float(m.group(7)),float(m.group(8)),
                                                                  float(m.group(2)),float(m.group(3)),float(m.group(4)),
                                                                  )))
            continue

        raise Exception("'transforms.txt': I only know about 'ins2veh' and 'cam2ins' lines. Got '{}'".
                        format(l))

    if not all(e is not None for e in x.values()):
        raise Exception("Transforms file '{}' incomplete. Missing values for: {}",
                        f.name,
                        [k for k in x.keys() if not x[k]])
    if needclose:
        f.close()

    return x




#!/usr/bin/env python3

r'''Checks the C and Python implementations of stereo_range()

This test needs external data, so it isn't included in the enabled-by-default set
'''

import sys
import numpy as np
import numpysane as nps
import os

testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import testutils
import cv2



# This test needs data external to this repository. Can be downloaded like this:
if False:
    import subprocess
    subprocess.check_output("wget -O /tmp/0.cameramodel https://mrcal.secretsauce.net/external/2022-11-05--dtla-overpass--samyang--alpha7/stereo/0.cameramodel && " +
                            "wget -O /tmp/1.cameramodel https://mrcal.secretsauce.net/external/2022-11-05--dtla-overpass--samyang--alpha7/stereo/1.cameramodel && "
                            "wget -O /tmp/0.jpg         https://mrcal.secretsauce.net/external/2022-11-05--dtla-overpass--samyang--alpha7/stereo/0.jpg && "
                            "wget -O /tmp/1.jpg         https://mrcal.secretsauce.net/external/2022-11-05--dtla-overpass--samyang--alpha7/stereo/1.jpg",
                            shell = True)

model_filenames=("/tmp/0.cameramodel",
                 "/tmp/1.cameramodel")
image_filenames=("/tmp/0.jpg",
                 "/tmp/1.jpg")

pixels_per_deg_az      = -.25
pixels_per_deg_el      = -.25
az_fov_deg             = dict(LENSMODEL_LATLON  = 160,
                              LENSMODEL_PINHOLE = 120)
el_fov_deg             = 60
az0_deg                = 0
el0_deg                = 0



try:
    models = [mrcal.cameramodel(f) \
              for f in model_filenames]
except FileNotFoundError:
    print("Data not found. Commands to download appear at the top of this script")
    sys.exit(1)

images = [mrcal.load_image(f, bits_per_pixel = 8, channels = 1) \
          for f in image_filenames]

clahe = cv2.createCLAHE()
clahe.setClipLimit(8)
images = [ clahe.apply(image) for image in images ]

# This is a hard-coded property of the OpenCV StereoSGBM implementation
disparity_scale = 16

q = np.array(((30,   100),
              (300,  200),
              (1000, 400),
              (500,  300),
              (80,   500),
              (400,  350),
              (900,  350),
              (1100, 350),
              (1400, 350),
              (400,  450),
              (900,  450),
              (1100, 450),
              (1400, 450),
              (1300, 450)))



for rectification_model in ('LENSMODEL_LATLON',
                            'LENSMODEL_PINHOLE',
                            ):

    models_rectified = \
        mrcal.rectified_system(models,
                               az_fov_deg          = az_fov_deg[rectification_model],
                               el_fov_deg          = el_fov_deg,
                               az0_deg             = az0_deg,
                               el0_deg             = el0_deg,
                               pixels_per_deg_az   = pixels_per_deg_az,
                               pixels_per_deg_el   = pixels_per_deg_el,
                               rectification_model = rectification_model)

    for disparity_min in (0,10,):
        disparity_max = 128

        # round to nearest multiple of disparity_scale. The OpenCV StereoSGBM
        # implementation requires this
        num_disparities = disparity_max - disparity_min
        num_disparities = disparity_scale*round(num_disparities/disparity_scale)

        stereo_sgbm = \
            cv2.StereoSGBM_create(minDisparity      = disparity_min,
                                  numDisparities    = num_disparities,
                                  P1                = 600,
                                  P2                = 2400,
                                  disp12MaxDiff     = 1,
                                  uniquenessRatio   = 5,
                                  speckleWindowSize = 100,
                                  speckleRange      = 2,
                                  mode              = cv2.StereoSGBM_MODE_SGBM)

        rectification_maps = mrcal.rectification_maps(models, models_rectified)

        images_rectified = [mrcal.transform_image(images[i],
                                                  rectification_maps[i]) \
                            for i in range(2)]

        disparity = stereo_sgbm.compute(*images_rectified)
        mask_valid = \
            (disparity > 0) * \
            (disparity >= disparity_min*disparity_scale) * \
            (disparity <= disparity_max*disparity_scale)

        ranges_dense = \
            mrcal.stereo_range( disparity,
                                models_rectified,
                                disparity_scale      = disparity_scale,
                                disparity_min        = disparity_min)
        mask_valid_ranges_dense = ranges_dense > 0

        ranges_dense_python = \
            mrcal.stereo._stereo_range_python( disparity,
                                               models_rectified,
                                               disparity_scale      = disparity_scale,
                                               disparity_min        = disparity_min)
        mask_valid_ranges_dense_pythion = ranges_dense_python > 0

        ranges_sparse = \
            mrcal.stereo_range( disparity[q[:,1], q[:,0]],
                                models_rectified,
                                qrect0               = q,
                                disparity_scale      = disparity_scale,
                                disparity_min        = disparity_min)

        ranges_sparse_python = \
            mrcal.stereo._stereo_range_python( disparity[q[:,1], q[:,0]],
                                               models_rectified,
                                               qrect0               = q,
                                               disparity_scale      = disparity_scale,
                                               disparity_min        = disparity_min)


        testutils.confirm_equal( ranges_dense,
                                 ranges_dense_python,
                                 worstcase = True,
                                 eps = 1e-3,
                                 msg=f'Dense stereo_range() matches in C and Python: rectification_model={rectification_model} disparity_min={disparity_min}')

        testutils.confirm_equal( ranges_sparse,
                                 ranges_sparse_python,
                                 worstcase = True,
                                 eps = 1e-3,
                                 msg=f'Sparse stereo_range() matches in C and Python: rectification_model={rectification_model} disparity_min={disparity_min}')

        testutils.confirm_equal( ranges_sparse,
                                 ranges_dense[q[:,1], q[:,0]],
                                 worstcase = True,
                                 eps = 1e-3,
                                 msg=f'Sparse and dense stereo_range() match: rectification_model={rectification_model} disparity_min={disparity_min}')

        testutils.confirm_equal( mask_valid_ranges_dense,
                                 mask_valid,
                                 msg=f'Dense stereo_range() invalid values handled correctly: rectification_model={rectification_model} disparity_min={disparity_min}')
        testutils.confirm_equal( mask_valid_ranges_dense_pythion,
                                 mask_valid,
                                 msg=f'Dense stereo_range() in Python: invalid values handled correctly: rectification_model={rectification_model} disparity_min={disparity_min}')
        testutils.confirm( np.all(ranges_dense[~mask_valid] == 0.),
                           msg=f'Dense stereo_range() values are all 0 in invalid areas: rectification_model={rectification_model} disparity_min={disparity_min}')



        if False:
            range_image_limits = (1,1000)
            disparity_colored = mrcal.apply_color_map(disparity,
                                                      a_min = disparity_min*disparity_scale,
                                                      a_max = disparity_max*disparity_scale)
            ranges_colored = mrcal.apply_color_map(ranges_dense,
                                                   a_min = range_image_limits[0],
                                                   a_max = range_image_limits[1])
            mrcal.save_image("/tmp/disparity.png", disparity_colored)
            mrcal.save_image("/tmp/range.png",     ranges_colored)
            print("Wrote /tmp/disparity.png and /tmp/range.png")

testutils.finish()



import sys
import numpy as np
import numpysane as nps
import copy
import os
import re

# I import the LOCAL mrcal since that's what I'm testing
testdir = os.path.dirname(os.path.realpath(__file__))
sys.path[:0] = f"{testdir}/..",
import mrcal

def sample_dqref(observations,
                 pixel_uncertainty_stdev,
                 make_outliers = False):

    # Outliers have weight < 0. The code will adjust the outlier observations
    # also. But that shouldn't matter: they're outliers so those observations
    # should be ignored
    weight  = observations[...,-1]
    q_noise = np.random.randn(*observations.shape[:-1], 2) * pixel_uncertainty_stdev / nps.mv(nps.cat(weight,weight),0,-1)

    if make_outliers:
        if not hasattr(sample_dqref, 'idx_outliers_ref_flat'):
            NobservedPoints = observations.size // 3
            sample_dqref.idx_outliers_ref_flat = \
                np.random.choice( NobservedPoints,
                                  (NobservedPoints//100,), # 1% outliers
                                  replace = False )
        nps.clump(q_noise, n=3)[sample_dqref.idx_outliers_ref_flat, :] *= 20

    observations_perturbed = observations.copy()
    observations_perturbed[...,:2] += q_noise
    return q_noise, observations_perturbed


def grad(f, x,
         *,
         switch              = None,
         forward_differences = False,
         step                = 1e-6):
    r'''Computes df/dx at x

    f is a function of one argument. If the input has shape Si and the output
    has shape So, the returned gradient has shape So+Si. This computes forward
    differences.

    If the function being evaluated produces bimodal output, the step might move
    us to a different mode, giving a falsely-high difference. Use the "switch"
    argument to switch to the other mode in this case.

    '''

    if switch is not None and not forward_differences:
        raise Exception("switch works ONLY with forward differences")

    d     = np.zeros(x.shape,dtype=float)
    dflat = d.ravel()

    def df_dxi(i, d,dflat):

        if not forward_differences:
            # central differences
            dflat[i] = step
            fplus  = f(x+d)
            fminus = f(x-d)
            j = (fplus-fminus)/(2.*step)
            dflat[i] = 0
            return j

        else:
            # forward differences
            dflat[i] = step
            f0    = f(x)
            fplus = f(x+d)
            if switch is not None and nps.norm2(fplus-f0) > 1.:
                fplus = switch(fplus)
            j = (fplus-f0)/step
            dflat[i] = 0
            return j

    # grad variable is in first dim
    Jflat = nps.cat(*[df_dxi(i, d,dflat) for i in range(len(dflat))])
    # grad variable is in last dim
    Jflat = nps.mv(Jflat, 0, -1)
    return Jflat.reshape( Jflat.shape[:-1] + d.shape )


def grad__r_from_R(R):

    # This function deals with a flattened Rflat with shape (9,)
    #
    # dr/dRflat isn't well-defined: not all sets of 9 numbers form a valid
    # rotation matrix: the r_from_R mapping locally acts in a 3-dimensional
    # subspace of R. We get this subspace from the opposite, well-defined
    # gradient dRflat/dr. dRflat_dr.shape is (9,3) so dR all lie in a subspace
    # spanned by its columns. I compute a QR decomposition:
    #
    #   dRflat_dr = Q R
    #
    # where Q is orthonormal (9,3) and R is some (3,3). Q is
    # the basis spanned by Rflat in response to perturbations in dr:
    #
    #   dRflat = Q R dr
    #
    # So for some dRflat I can project to the subspace and transform to get dr:
    #
    #   dr = inv(R) Qt dRflat
    #
    # This is a different way of computing the same thing that
    # r_from_R(get_gradients=True) does
    #
    # Note that I'm not computing any numerical gradients here. If R_from_r() is
    # right, then this should be right too
    r = mrcal.r_from_R(R)

    _,dR_dr      = mrcal.R_from_r(r, get_gradients = True)
    dRflat_dr    = nps.clump(dR_dr, n=2)

    Q,R = np.linalg.qr(dRflat_dr)
    return np.linalg.solve(R, Q.T)


def calibration_baseline(model, Ncameras, Nframes, extra_observation_at,
                         object_width_n,
                         object_height_n,
                         object_spacing,
                         # Camera poses in respect to the first camera. Thus
                         # rt_cam_ref_true[0] = zeros(6). This is
                         # checked
                         rt_cam_cam0_true,
                         calobject_warp_true,
                         testdir,
                         *,
                         cull_left_of_center = False,
                         range_to_boards = 4.0,
                         x_offset = 0.0,
                         x_noiseradius = 2.5,
                         y_noiseradius = 2.5,
                         x_mirror      = False, # half with +x_offset, half with -x_offset
                         points        = False,

                         moving_cameras     = False,
                         ref_frame0         = False,
                         do_optimize_frames = True,

                         report_vanilla_and_requested = False,
                         optimize_models = True,

                         # The logic to avoid oblique views was added in
                         #   https://github.com/dkogan/mrcal/commit/b54df5d3
                         #
                         # We want to disable this if we're trying to compare
                         # results to mrcal 2.4, since this commit was made
                         # after that
                         avoid_oblique_views = True):

    r'''Compute a calibration baseline as a starting point for experiments

This is a perfect, noiseless solve. Regularization IS enabled, and the returned
model is at the optimization optimum. So the returned models will not sit
exactly at the ground-truth.

If x_mirror then half of Nframes are shifted in the x direction by x_offset and
the other half by -x_offset

I always compute chessboard views. But if points: I store each corner
observation as a separate point. if points: I ALSO enable the unity_cam01
regularization to make the solve non-singular


The moving_cameras and ref_frame0 arguments control what is moving/stationary
and where the reference coordinate system is. Since in the end we look at
relative camera-frame transforms, their absolute poses don't matter, and either
the cameras or frames can be stationary. And the reference coord system can be
defined at either one of the cameras or one of the frames. Some of these
combinations aren't supported at all, and some are supported only if we have a
single camera. An exception will be raised for an un-supported case. See the
docstring for test/test-projection-uncertainty.py for notes about the cases and
what is supported

ARGUMENTS

- model: string. 'opencv4' or 'opencv8' or 'splined'

- ...

    '''


    if nps.norm2(rt_cam_cam0_true[0]) > 0:
        raise Exception("A non-identity cam0 transform was given. This is not supported")

    if re.match('opencv',model):
        models_true_refcam0 = ( mrcal.cameramodel(f"{testdir}/data/cam0.opencv8.cameramodel"),
                                mrcal.cameramodel(f"{testdir}/data/cam0.opencv8.cameramodel"),
                                mrcal.cameramodel(f"{testdir}/data/cam1.opencv8.cameramodel"),
                                mrcal.cameramodel(f"{testdir}/data/cam1.opencv8.cameramodel") )

        if model == 'opencv4':
            # I have opencv8 models_true_refcam0, but I truncate to opencv4 models_true_refcam0
            for m in models_true_refcam0:
                m.intrinsics( intrinsics = ('LENSMODEL_OPENCV4', m.intrinsics()[1][:8]))
    elif model == 'splined':
        models_true_refcam0 = ( mrcal.cameramodel(f"{testdir}/data/cam0.splined.cameramodel"),
                                mrcal.cameramodel(f"{testdir}/data/cam0.splined.cameramodel"),
                                mrcal.cameramodel(f"{testdir}/data/cam1.splined.cameramodel"),
                                mrcal.cameramodel(f"{testdir}/data/cam1.splined.cameramodel") )
    else:
        raise Exception("Unknown lens being tested")

    models_true_refcam0 = models_true_refcam0[:Ncameras]
    lensmodel           = models_true_refcam0[0].intrinsics()[0]
    Nintrinsics         = mrcal.lensmodel_num_params(lensmodel)

    for i in range(Ncameras):
        models_true_refcam0[i].rt_cam_ref(rt_cam_cam0_true[i])

    imagersizes = nps.cat( *[m.imagersize() for m in models_true_refcam0] )

    # These are perfect
    intrinsics_true = nps.cat( *[m.intrinsics()[1] for m in models_true_refcam0] )
    x_center = -(Ncameras-1)/2.



    def synthesize(z, z_noiseradius, Nframes,
                   x_offset):
        if avoid_oblique_views:
            kwargs = \
                dict( pcamera_nominal_ref   = np.array((x_center,0,0), dtype=float),
                      max_oblique_angle_deg = 30. )
        else:
            kwargs = dict()


        return \
            mrcal.synthesize_board_observations(models_true_refcam0,
                                                object_width_n  = object_width_n,
                                                object_height_n = object_height_n,
                                                object_spacing  = object_spacing,
                                                calobject_warp  = calobject_warp_true,
                                                Nframes         = Nframes,
                                                rt_ref_boardcenter = np.array((0.,
                                                                               0.,
                                                                               0.,
                                                                               x_center + x_offset,
                                                                               0,
                                                                               z)),
                                                rt_ref_boardcenter__noiseradius = np.array((np.pi/180.*30.,
                                                                                            np.pi/180.*30.,
                                                                                            np.pi/180.*20.,
                                                                                            x_noiseradius,
                                                                                            y_noiseradius,
                                                                                            z_noiseradius)),
                                                **kwargs)


    if not x_mirror:
        # shapes (Nframes, Ncameras, Nh, Nw, 2),
        #        (Nframes, 4,3)
        q_true,Rt_cam0_board_true = \
            synthesize(z             = range_to_boards,
                       z_noiseradius = range_to_boards / 2.0,
                       Nframes       = Nframes,
                       x_offset      = x_offset)
    else:
        if Nframes & 1:
            raise Exception("x_mirror is True, so Nframes must be even")

        # shapes (Nframes//2, Ncameras, Nh, Nw, 2),
        #        (Nframes//2, 4,3)
        q_true0,Rt_cam0_board_true0 = \
            synthesize(z             = range_to_boards,
                       z_noiseradius = range_to_boards / 2.0,
                       Nframes       = Nframes//2,
                       x_offset      = x_offset)
        q_true1,Rt_cam0_board_true1 = \
            synthesize(z             = range_to_boards,
                       z_noiseradius = range_to_boards / 2.0,
                       Nframes       = Nframes//2,
                       x_offset      = -x_offset)

        q_true            = nps.glue(q_true0,           q_true1,            axis=-5)
        Rt_cam0_board_true = nps.glue(Rt_cam0_board_true0,Rt_cam0_board_true1, axis=-3)


    if extra_observation_at is not None:
        q_true_extra,Rt_cam0_board_true_extra = \
            synthesize(z             = extra_observation_at,
                       z_noiseradius = extra_observation_at / 10.0,
                       Nframes       = 1,
                       x_offset      = x_offset)


        q_true            = nps.glue( q_true, q_true_extra,
                                      axis=-5)
        Rt_cam0_board_true = nps.glue( Rt_cam0_board_true, Rt_cam0_board_true_extra,
                                      axis=-3)

        Nframes += 1

    # shape (Nframes, 6)
    rt_cam0_board_true = mrcal.rt_from_Rt(Rt_cam0_board_true)

    ############# I have perfect observations in q_true.
    # weight has shape (Nframes, Ncameras, Nh, Nw),
    weight01 = (np.random.rand(*q_true.shape[:-1]) + 1.) / 2. # in [0,1]
    weight0 = 0.2
    weight1 = 1.0
    weight = weight0 + (weight1-weight0)*weight01

    if cull_left_of_center:

        imagersize = models_true_refcam0[0].imagersize()
        for m in models_true_refcam0[1:]:
            if np.any(m.imagersize() - imagersize):
                raise Exception("I'm assuming all cameras have the same imager size, but this is false")

        weight[q_true[...,0] < imagersize[0]/2.] /= 1000.

    # I want observations of shape (Nframes*Ncameras, Nh, Nw, 3) where each row is
    # (x,y,weight)
    observations_board_true = \
        nps.clump( nps.glue(q_true,
                            nps.dummy(weight,-1),
                            axis=-1),
                   n=2)

    # Dense observations. All the cameras see all the boards
    indices_frame_camera = np.zeros( (Nframes,Ncameras,2), dtype=np.int32)
    indices_frame_camera[...,0] += nps.transpose(np.arange(Nframes, dtype=np.int32))
    indices_frame_camera[...,1] += np.arange(Ncameras, dtype=np.int32)
    # shape (Nframes*Ncameras, 2)
    indices_frame_camera = nps.clump(indices_frame_camera, n=2)

    # stationary cameras, so idxci = idxce
    # shape (Nframes*Ncameras, 3)
    indices_frame_camintrinsics_camextrinsics = \
        nps.glue(indices_frame_camera,
                 indices_frame_camera[:,(1,)] - 1,
                 axis=-1)

    # I now reoptimize the perfect-observations problem. Without regularization,
    # this is a no-op: I'm already at the optimum. With regularization, this will
    # move us a certain amount (that the test will evaluate). Then I look at
    # noise-induced motions off this optimization optimum
    optimization_inputs_baseline = \
        dict( intrinsics                                = copy.deepcopy(intrinsics_true),
              rt_cam_ref                                = copy.deepcopy(rt_cam_cam0_true[1:,:]),
              rt_ref_frame                              = copy.deepcopy(rt_cam0_board_true),
              points                                    = None,
              observations_board                        = observations_board_true,
              indices_frame_camintrinsics_camextrinsics = indices_frame_camintrinsics_camextrinsics,
              observations_point                        = None,
              indices_point_camintrinsics_camextrinsics = None,
              lensmodel                                 = lensmodel,
              calobject_warp                            = copy.deepcopy(calobject_warp_true),
              imagersizes                               = imagersizes,
              calibration_object_spacing                = object_spacing,
              verbose                                   = False,
              # I add this later, in _apply_moving_ref()
              do_optimize_frames                        = True,
              do_optimize_intrinsics_core               = False if model =='splined' else True,
              do_optimize_intrinsics_distortions        = True,
              do_optimize_extrinsics                    = True,
              do_optimize_calobject_warp                = True,
              do_apply_regularization                   = True,
              do_apply_outlier_rejection                = False,
              do_apply_regularization_unity_cam01       = points)

    ###########################################################################
    # p = mrcal.show_geometry(models_true_refcam0,
    #                         frames          = rt_cam0_board_true,
    #                         object_width_n  = object_width_n,
    #                         object_height_n = object_height_n,
    #                         object_spacing  = object_spacing)
    # sys.exit()

    if report_vanilla_and_requested:
        optimization_inputs_vanilla_baseline = copy.deepcopy(optimization_inputs_baseline)
        if optimize_models:
            mrcal.optimize(**optimization_inputs_vanilla_baseline)

        optimization_inputs_list = \
            (optimization_inputs_vanilla_baseline,
             optimization_inputs_baseline)
    else:
        optimization_inputs_list = \
            (optimization_inputs_baseline,)


    _apply_moving_ref(optimization_inputs_baseline,
                      moving_cameras     = moving_cameras,
                      ref_frame0         = ref_frame0,
                      do_optimize_frames = do_optimize_frames)
    if points:
        _calibration_boards_to_points(optimization_inputs_baseline)
        points_true = np.array(optimization_inputs_baseline['points'])
        rt_ref_board_true = None
    else:
        rt_ref_board_true = np.array(optimization_inputs_baseline['rt_ref_frame'])
    rt_cam_ref_true = np.array(optimization_inputs_baseline['rt_cam_ref'])

    if optimize_models:
        mrcal.optimize(**optimization_inputs_baseline)


    if not points:
        return ( *optimization_inputs_list,                \
                 models_true_refcam0,                      \
                 rt_cam_ref_true,                          \
                 rt_ref_board_true )

    else:
        return ( *optimization_inputs_list,                \
                 models_true_refcam0,                      \
                 rt_cam_ref_true,                          \
                 points_true )


def _apply_moving_ref(optimization_inputs,
                      *,
                      moving_cameras,
                      ref_frame0,
                      do_optimize_frames):
    r'''See the docstring for test-projection-uncertainty.py for a description
    of all the cases'''

    if not moving_cameras and not ref_frame0:
        if do_optimize_frames:
            # This is the baseline case. The data is already set up like this.
            return

        # Frames are fixed. This is the old "fixedframes" mode. I allow cam0 to
        # move
        optimization_inputs['indices_frame_camintrinsics_camextrinsics'][:,2] += 1
        optimization_inputs['rt_cam_ref'] = nps.glue( np.zeros((6,), dtype=float),
                                                      optimization_inputs['rt_cam_ref'],
                                                      axis = -2 )
        optimization_inputs['do_optimize_extrinsics'] = True
        optimization_inputs['do_optimize_frames'    ] = do_optimize_frames
        return

    idxf,idxci,idxce   = optimization_inputs['indices_frame_camintrinsics_camextrinsics'].T
    Ncameras           = len(optimization_inputs['intrinsics'])
    rt_cam0_board_true = optimization_inputs['rt_ref_frame']

    if moving_cameras and \
       ref_frame0:

        if Ncameras > 1:
            raise Exception("Scenario (moving-camera, ref-at-frame0) cannot work with multiple cameras: camera-rig implementation is required")
        if do_optimize_frames:
            raise Exception("Scenario (moving-camera, ref-at-frame0) requires !do_optimize_frames")

        optimization_inputs['indices_frame_camintrinsics_camextrinsics'] = \
            np.ascontiguousarray(nps.transpose( nps.cat( idxce+1,
                                                         idxci,
                                                         idxf ) ))
        optimization_inputs['rt_cam_ref'            ] = np.array(rt_cam0_board_true)
        optimization_inputs['rt_ref_frame'          ] = np.zeros((1,6,), dtype=float)
        optimization_inputs['do_optimize_extrinsics'] = True
        optimization_inputs['do_optimize_frames'    ] = do_optimize_frames
        return


    if moving_cameras and \
       not ref_frame0:

        if Ncameras > 1:
            raise Exception("Scenario (moving-camera, ref-at-cam0) cannot work with multiple cameras: camera-rig implementation is required")

        optimization_inputs['indices_frame_camintrinsics_camextrinsics'] = \
            np.ascontiguousarray(nps.transpose( nps.cat( idxce+1, # all 0: I have one board, and it can move
                                                         idxci,
                                                         idxf-1 )))
        rt_cam_cam0 = \
            mrcal.compose_rt(rt_cam0_board_true[1:,:],
                             mrcal.invert_rt(rt_cam0_board_true[0,:]))
        optimization_inputs['rt_cam_ref'            ] = rt_cam_cam0
        optimization_inputs['rt_ref_frame'          ] = rt_cam0_board_true[(0,),:]
        optimization_inputs['do_optimize_extrinsics'] = True
        optimization_inputs['do_optimize_frames'    ] = do_optimize_frames
        return


    if not moving_cameras and \
       ref_frame0:
        raise Exception("Case not supported: not moving_cameras and ref_frame0")

    raise Exception("Unhandled case. Getting here is a bug")


def _calibration_boards_to_points(optimization_inputs):

    # I break up the chessboard observations into discrete points

    rt_ref_board                              = optimization_inputs['rt_ref_frame']
    Rt_ref_board                              = mrcal.Rt_from_rt(rt_ref_board)
    Nframes                                   = len(rt_ref_board)
    indices_frame_camintrinsics_camextrinsics = optimization_inputs['indices_frame_camintrinsics_camextrinsics']
    observations_board                        = optimization_inputs['observations_board']
    Nobservations                             = len(observations_board)
    object_height_n,object_width_n            = observations_board.shape[-3:-1]
    object_spacing                            = optimization_inputs['calibration_object_spacing']
    calobject_warp                            = optimization_inputs['calobject_warp']

    # shape (Nobservations,H,W,3)
    indices_point_camintrinsics_camextrinsics = np.zeros((Nobservations,object_height_n,object_width_n,3), dtype=np.int32)

    # shape (Nobservations,)
    iframe = indices_frame_camintrinsics_camextrinsics[...,0]

    # index_point
    # shape (Nobservations,H,W)
    indices_point_camintrinsics_camextrinsics[...,0] = \
        np.arange(object_height_n * object_width_n).reshape(object_height_n,object_width_n) + \
        nps.dummy(iframe * object_height_n * object_width_n,
                  -1,-1)

    # camintrinsics and camextrinsics
    # shape (Nobservations,H,W,2)
    indices_point_camintrinsics_camextrinsics[...,1:] = \
        nps.dummy(indices_frame_camintrinsics_camextrinsics[...,1:], -2,-2)

    # shape (Nobservations*H*W, 3)
    indices_point_camintrinsics_camextrinsics = \
        nps.clump(indices_point_camintrinsics_camextrinsics, n=3)

    # shape (H,W,3)
    pboard = \
        mrcal.ref_calibration_object(object_width_n,
                                     object_height_n,
                                     object_spacing,
                                     calobject_warp = calobject_warp)

    # shape (Nframes,1,1,4,3)
    Rt_ref_board_frames = nps.dummy(Rt_ref_board,-3,-3)

    # shape (Nframes,H,W, 3)
    pref = mrcal.transform_point_Rt(Rt_ref_board_frames,
                                    pboard)
    # shape (Nframes*H*W, 3)
    pref = nps.clump(pref, n=3)

    # shape (Nobservations*Nh*Nw, 3)
    observations_point = nps.clump(observations_board, n=3)

    # At this point the data is conceptually good. However, an implementation
    # detail requires monotonically non-decreasing point indices:
    #   uncertainty.c(983): Unexpected jacobian structure. I'm assuming
    #   non-decreasing point references. The Jcross_t__Jcross computation uses
    #   chunks of Kpackedp; it assumes that once the chunk is computed, it is
    #   DONE, and never revisited. Non-monotonic point indices break that
    # So I re-sort here
    iipoint = np.argsort(indices_point_camintrinsics_camextrinsics[:,0],
                         kind='stable')
    indices_point_camintrinsics_camextrinsics = \
        indices_point_camintrinsics_camextrinsics[iipoint]
    observations_point = observations_point[iipoint]


    optimization_inputs['indices_point_camintrinsics_camextrinsics'] = indices_point_camintrinsics_camextrinsics
    optimization_inputs['points']                                    = pref
    optimization_inputs['observations_point']                        = observations_point

    optimization_inputs['indices_frame_camintrinsics_camextrinsics'] = None
    optimization_inputs['rt_ref_frame']                              = None
    optimization_inputs['observations_board']                        = None

    # point-only solves are unique up-to-scale only, and need the extra regularization
    Ncameras_extrinsics = optimization_inputs['rt_cam_ref'].shape[0]
    if Ncameras_extrinsics <= 1:
        if optimization_inputs['do_optimize_frames']:
            raise Exception("Points-only solves with do_optimize_frames==True need extra regularization, but currently that's only possible with Ncameras_extrinsics>=1")
    else:
        optimization_inputs['do_apply_regularization_unity_cam01'] = True
        # I rescale all geometry to make the unity01 regularization=0 at the start
        s = 1. / nps.norm2(optimization_inputs['rt_cam_ref'][0,3:])
        optimization_inputs['rt_cam_ref'][:,3:] *= s
        optimization_inputs['points'] *= s


def calibration_sample(Nsamples,
                       optimization_inputs_baseline,
                       pixel_uncertainty_stdev,
                       function_optimize = None):

    r'''Sample calibrations subject to random noise on the input observations

optimization_inputs_baseline['observations_board'] and
optimization_inputs_baseline['observations_point'] are assumed to contain
perfect observations

    '''

    def have(k):
        return k in optimization_inputs_baseline and \
            optimization_inputs_baseline[k] is not None

    intrinsics_sampled = np.zeros((Nsamples,) + optimization_inputs_baseline['intrinsics']    .shape, dtype=float)

    if have('rt_ref_frame'):
        frames_sampled = np.zeros((Nsamples,) + optimization_inputs_baseline['rt_ref_frame'].shape, dtype=float)
        q_noise_board_sampled = np.zeros((Nsamples,) + optimization_inputs_baseline['observations_board'].shape[:-1] + (2,), dtype=float)
    else:
        frames_sampled        = None
        q_noise_board_sampled = None
    if have('points'):
        points_sampled = np.zeros((Nsamples,) + optimization_inputs_baseline['points'].shape, dtype=float)
        q_noise_point_sampled = np.zeros((Nsamples,) + optimization_inputs_baseline['observations_point'].shape[:-1] + (2,), dtype=float)
    else:
        points_sampled        = None
        q_noise_point_sampled = None
    if have('calobject_warp'):
        calobject_warp_sampled = np.zeros((Nsamples,) + optimization_inputs_baseline['calobject_warp'].shape, dtype=float)
    else:
        calobject_warp_sampled = None

    if function_optimize is None:
        b_sampled_unpacked = np.zeros((Nsamples,mrcal.num_states(**optimization_inputs_baseline)),)
    else:
        b_sampled_unpacked = None

    optimization_inputs_sampled = [None] * Nsamples

    have_fixed_cameras = \
        ( have('indices_frame_camintrinsics_camextrinsics') and \
          np.any(optimization_inputs_baseline['indices_frame_camintrinsics_camextrinsics'][:,2] < 0)) or \
        ( have('indices_point_camintrinsics_camextrinsics') and \
          np.any(optimization_inputs_baseline['indices_point_camintrinsics_camextrinsics'][:,2] < 0))

    Ncameras_extrinsics = optimization_inputs_baseline['rt_cam_ref'].shape[0]
    if have_fixed_cameras:
        # the first row is fixed at 0
        Ncameras_extrinsics += 1
    extrinsics_sampled_mounted = np.zeros((Nsamples,Ncameras_extrinsics,6), dtype=float)

    for isample in range(Nsamples):
        if (isample+1) % 20 == 0:
            print(f"Sampling {isample+1}/{Nsamples}")

        optimization_inputs_sampled[isample] = copy.deepcopy(optimization_inputs_baseline)
        optimization_inputs = optimization_inputs_sampled[isample]

        if have('observations_board'):
            q_noise_board_sampled[isample],optimization_inputs['observations_board'] = \
                sample_dqref(optimization_inputs['observations_board'],
                             pixel_uncertainty_stdev)
        else:
            q_noise_board = None
        if have('observations_point'):
            q_noise_point_sampled[isample],optimization_inputs['observations_point'] = \
                sample_dqref(optimization_inputs['observations_point'],
                             pixel_uncertainty_stdev)
        else:
            q_noise_point = None

        if function_optimize is None:
            b_sampled_unpacked[isample] = mrcal.optimize(**optimization_inputs)['b_packed']
            mrcal.unpack_state(b_sampled_unpacked[isample], **optimization_inputs)
        else:
            function_optimize(optimization_inputs)

        intrinsics_sampled    [isample,...] = optimization_inputs['intrinsics']
        if not have_fixed_cameras:
            extrinsics_sampled_mounted[isample,   ...] = optimization_inputs['rt_cam_ref']
        else:
            # the remaining row is already 0
            extrinsics_sampled_mounted[isample,1:,...] = optimization_inputs['rt_cam_ref']

        if frames_sampled is not None:
            frames_sampled[isample,...] = optimization_inputs['rt_ref_frame']
        if points_sampled is not None:
            points_sampled[isample,...] = optimization_inputs['points']
        if calobject_warp_sampled is not None:
            calobject_warp_sampled[isample,...] = optimization_inputs['calobject_warp']


    return (intrinsics_sampled,
            extrinsics_sampled_mounted,
            frames_sampled,
            points_sampled,
            calobject_warp_sampled,
            q_noise_board_sampled,
            q_noise_point_sampled,
            b_sampled_unpacked,
            optimization_inputs_sampled)



#!/usr/bin/env python3

r'''Test of the save_image() and load_image() functions
'''

import sys
import numpy as np
import numpysane as nps
import os

testdir = os.path.dirname(os.path.realpath(__file__))

# I import the LOCAL mrcal since that's what I'm testing
sys.path[:0] = f"{testdir}/..",
import mrcal
import testutils



import tempfile
import atexit
import shutil
workdir = tempfile.mkdtemp()
def cleanup():
    global workdir
    try:
        shutil.rmtree(workdir)
        workdir = None
    except:
        pass
atexit.register(cleanup)


W,H = 133,211

images = dict()
images[8] = np.arange(W*H, dtype=np.uint8).reshape(H,W)
images[8] *= images[8]

images[16] = np.arange(W*H, dtype=np.uint16).reshape(H,W)
images[16] *= images[16]

images[24] = np.arange(W*H*3, dtype=np.uint8).reshape(H,W,3)
images[24] *= images[24]

filename_pattern = f"{workdir}/tst.{{extension}}"


def _check_load(filename, image_ref, what,
                should_fail_load    = False,
                compare_value       = True,
                extension           = None,
                **kwargs):

    if should_fail_load:
        testutils.confirm_raises(lambda: mrcal.load_image(filename, **kwargs),
                                 msg = f"Loading {what} image should fail")
        return


    try:
        image_check = mrcal.load_image(filename, **kwargs)
    except Exception as e:
        testutils.confirm(False,
                          msg = f"Error loading {what} image: {e}")

        return
    testutils.confirm(True,
                      msg = f"Success loading {what} image")

    # print(image_ref.shape)
    # print(image_check.shape)
    # print(image_ref.dtype)
    # print(image_check.dtype)

    bpp_load      = kwargs.get('bits_per_pixel')
    channels_load = kwargs.get('channels')
    if bpp_load is not None:
        H,W = image_ref.shape[:2]
        if bpp_load == 8:
            testutils.confirm_equal(image_check.shape, (H,W),
                                    msg = "Shapes match. Loading 8bpp-1channel image")
            testutils.confirm_equal(image_check.ndim, 2,
                                    msg = "channels match. Loading 8bpp-1channel image")
            testutils.confirm(image_check.dtype == np.uint8,
                              msg = "dtype match. Loading 8bpp-1channel image")
        elif bpp_load == 16:
            testutils.confirm_equal(image_check.shape, (H,W),
                                    msg = "Shapes match. Loading 16bpp-1channel image")
            testutils.confirm_equal(image_check.ndim, 2,
                                    msg = "channels match. Loading 16bpp-1channel image")
            testutils.confirm(image_check.dtype == np.uint16,
                              msg = "dtype match. Loading 16bpp-1channel image")
        elif bpp_load == 24:
            testutils.confirm_equal(image_check.shape, (H,W,3),
                                    msg = "Shapes match. Loading 24bpp-3channel image")
            testutils.confirm_equal(image_check.ndim, 3,
                                    msg = "channels match. Loading 24bpp-3channel image")
            testutils.confirm(image_check.dtype == np.uint8,
                              msg = "dtype match. Loading 24bpp-3channel image")

    if compare_value:
        # jpg is approximate
        if extension == 'jpg':
            eps = 10
            percentile = 95
            worstcase  = False
        else:
            eps = 1e-6
            percentile = None
            worstcase  = True

        if bpp_load == 24 and image_ref.ndim==2:
            image_ref = nps.glue( nps.dummy(image_ref,-1),
                                  nps.dummy(image_ref,-1),
                                  nps.dummy(image_ref,-1),
                                  axis = -1)
        # signed cast to make approximate comparisons work. Otherwise <0 wraps
        # around
        testutils.confirm_equal(image_ref.astype(np.int32), image_check.astype(np.int32),
                                worstcase  = worstcase,
                                eps        = eps,
                                percentile = percentile,
                                msg = f"load/save match for {what}")





for bpp,extension,what in ( (8,  'png', "8bpp  png grayscale"),
                            (16, 'png', "16bpp png grayscale"),
                            (24, 'png', "24bpp png bgr"),
                            (8,  'jpg', "8bpp  jpg grayscale"),
                            (24, 'jpg', "24bpp jpg bgr"),
                            (8,  'pgm', "8bpp  pgm grayscale"),
                            (16, 'pgm', "16bpp pgm grayscale"),
                            (24, 'ppm', "24bpp ppm bgr"),
                           ):

    filename = filename_pattern.format(extension=extension)

    image = images[bpp]

    try:
        mrcal.save_image(filename, image)
    except:
        testutils.confirm(False,
                          msg = f"Error saving {what} image")
        continue
    testutils.confirm(True,
                      msg = f"Success saving {what} image")

    def check_load(**kwargs):
        if 'bits_per_pixel' not in kwargs:
            _check_load(filename, image,
                        what,
                        extension = extension,
                        **kwargs)
        else:
            _check_load(filename, image,
                        f"{what} (loading as {kwargs['bits_per_pixel']}bpp)",
                        extension = extension,
                        **kwargs)

    check_load()

    if bpp == 8:
        check_load(bits_per_pixel = 16,
                   channels       = 1,
                   should_fail_load = True)
        check_load(bits_per_pixel = 24,
                   channels       = 3)

    if bpp == 16:
        check_load(bits_per_pixel = 8,
                   channels       = 1,
                   # values won't match: this path equalizes
                   compare_value  = False)
        check_load(bits_per_pixel = 24,
                   channels       = 3,
                   should_fail_load = True)

    if bpp == 24:
        check_load(bits_per_pixel = 8,
                   channels       = 1,
                   # values won't match: colors are collapsed
                   compare_value  = False)
        check_load(bits_per_pixel = 16,
                   channels       = 1,
                   should_fail_load = True)

testutils.finish()
